import py.test
from ._tools import *


class DummyPreserveDimNode(mdp.PreserveDimNode):
    """Non-abstract dummy version of PreserveDimNode."""
    
    def is_trainable(self):
        return False
    

def testPreserveDimNode():
    """Test the different dimension setting options."""
    dim = 3
    node = DummyPreserveDimNode(input_dim=dim, output_dim=dim)
    assert node.output_dim == dim
    assert node.input_dim == dim
    node = DummyPreserveDimNode(input_dim=3)
    assert node.output_dim == dim
    assert node.input_dim == dim
    node = DummyPreserveDimNode(output_dim=3)
    assert node.output_dim == dim
    assert node.input_dim == dim
    node = DummyPreserveDimNode(output_dim=3)
    node.input_dim = dim
    assert node.output_dim == dim
    assert node.input_dim == dim
    def get_node():
        DummyPreserveDimNode(input_dim=dim, output_dim=dim+1)
    py.test.raises(mdp.InconsistentDimException, get_node)



# -*- coding:utf-8; -*-
"""Wraps the algorithms defined in scikits.learn in MDP Nodes.

This module is based on the 0.6.X branch of scikits.learn .
"""
from __future__ import print_function
__docformat__ = "restructuredtext en"

try:
    import sklearn
    _sklearn_prefix = 'sklearn'
except ImportError:
    import scikits.learn as sklearn
    _sklearn_prefix = 'scikits.learn'

import inspect
import re

import mdp

class ScikitsException(mdp.NodeException):
    """Base class for exceptions in nodes wrapping scikits algorithms."""
    pass

# import all submodules of sklearn (to work around lazy import)
from mdp.configuration import _version_too_old
if _version_too_old(sklearn.__version__, (0, 8)):
    scikits_modules = ['ann', 'cluster', 'covariance', 'feature_extraction',
                       'feature_selection', 'features', 'gaussian_process', 'glm',
                       'linear_model', 'preprocessing', 'svm',
                       'pca', 'lda', 'hmm', 'fastica', 'grid_search', 'mixture',
                       'naive_bayes', 'neighbors', 'qda']
elif _version_too_old(sklearn.__version__, (0, 9)):
    # package structure has been changed in 0.8
    scikits_modules = ['svm', 'linear_model', 'naive_bayes', 'neighbors',
                       'mixture', 'hmm', 'cluster', 'decomposition', 'lda',
                       'covariance', 'cross_val', 'grid_search',
                       'feature_selection.rfe', 'feature_extraction.image',
                       'feature_extraction.text', 'pipelines', 'pls',
                       'gaussian_process', 'qda']
elif _version_too_old(sklearn.__version__, (0, 11)):
    # from release 0.9 cross_val becomes cross_validation and hmm is deprecated
    scikits_modules = ['svm', 'linear_model', 'naive_bayes', 'neighbors',
                       'mixture', 'cluster', 'decomposition', 'lda',
                       'covariance', 'cross_validation', 'grid_search',
                       'feature_selection.rfe', 'feature_extraction.image',
                       'feature_extraction.text', 'pipelines', 'pls',
                       'gaussian_process', 'qda', 'ensemble', 'manifold',
                       'metrics', 'preprocessing', 'tree']    
elif _version_too_old(sklearn.__version__, (0, 17)):
    scikits_modules = ['svm', 'linear_model', 'naive_bayes', 'neighbors',
                       'mixture', 'cluster', 'decomposition', 'lda',
                       'covariance', 'cross_validation', 'grid_search',
                       'feature_selection', 'feature_extraction',
                       'pipeline', 'pls', 'gaussian_process', 'qda',
                       'ensemble', 'manifold', 'metrics', 'preprocessing',
                       'semi_supervised', 'tree', 'hmm']

else:
    scikits_modules = ['calibration', 'cluster', 'covariance', 'cross_decomposition',
                       'cross_validation', 'decomposition', 'discriminant_analysis',
                       'ensemble', 'feature_extraction', 'feature_selection',
                       'gaussian_process', 'grid_search', 'isotonic', 'kernel_approximation',
                       'kernel_ridge', 'learning_curve', 'linear_model', 'manifold',
                       'metrics', 'mixture', 'multiclass', 'naive_bayes', 'neighbors',
                       'neural_network', 'preprocessing', 'random_projection',
                       'semi_supervised', 'svm', 'tree']


for name in scikits_modules:
    # not all modules may be available due to missing dependencies
    # on the user system.
    # we just ignore failing imports
    try:
        __import__(_sklearn_prefix + '.' + name)
    except ImportError:
        pass


_WS_LINE_RE = re.compile(r'^\s*$')
_WS_PREFIX_RE = re.compile(r'^(\s*)')
_HEADINGS_RE = re.compile(r'''^(Parameters|Attributes|Methods|Examples|Notes)\n
                           (----+|====+)''', re.M + re.X)
_UNDERLINE_RE = re.compile(r'----+|====+')
_VARWITHUNDER_RE = re.compile(r'(\s|^)([a-zA-Z_][a-zA-Z0-9_]*_)(\s|$|[,.])')

_HEADINGS = set(['Parameters', 'Attributes', 'Methods', 'Examples',
                 'Notes', 'References'])

_DOC_TEMPLATE = """
%s

This node has been automatically generated by wrapping the ``%s.%s`` class
from the ``sklearn`` library.  The wrapped instance can be accessed
through the ``scikits_alg`` attribute.

%s
"""

def _gen_docstring(object, docsource=None):
    module = object.__module__
    name = object.__name__
    if docsource is None:
        docsource = object
    docstring = docsource.__doc__
    if docstring is None:
        return None

    lines = docstring.strip().split('\n')
    for i,line in enumerate(lines):
        if _WS_LINE_RE.match(line):
            break
    header = [line.strip() for line in lines[:i]]

    therest = [line.rstrip() for line in lines[i+1:]]
    body = []

    if therest:
        prefix = min(len(_WS_PREFIX_RE.match(line).group(1))
                     for line in therest if line)
        quoteind = None
        for i, line in enumerate(therest):
            line = line[prefix:]
            if line in _HEADINGS:
                body.append('**%s**' % line)
            elif _UNDERLINE_RE.match(line):
                body.append('')
            else:
                line = _VARWITHUNDER_RE.sub(r'\1``\2``\3', line)
                if quoteind:
                    if len(_WS_PREFIX_RE.match(line).group(1)) >= quoteind:
                        line = quoteind * ' ' + '- ' + line[quoteind:]
                    else:
                        quoteind = None
                        body.append('')
                body.append(line)

            if line.endswith(':'):
                body.append('')
                if i+1 < len(therest):
                    next = therest[i+1][prefix:]
                    quoteind = len(_WS_PREFIX_RE.match(next).group(1))

    return _DOC_TEMPLATE % ('\n'.join(header), module, name, '\n'.join(body))

# TODO: generalize dtype support
# TODO: have a look at predict_proba for Classifier.prob
# TODO: inverse <-> generate/rvs
# TODO: deal with input_dim/output_dim
# TODO: change signature of overwritten functions
# TODO: wrap_scikits_instance
# TODO: add sklearn availability to test info strings
# TODO: which tests ? (test that particular algorithm are / are not trainable)
# XXX: if class defines n_components, allow output_dim, otherwise throw exception
#      also for classifiers (overwrite _set_output_dim)
#      Problem: sometimes they call it 'k' (e.g., algorithms in sklearn.cluster)

def apply_to_scikits_algorithms(current_module, action,
                                processed_modules=None,
                                processed_classes=None):
    """ Function that traverses a module to find scikits algorithms.

    'sklearn' algorithms are identified by the 'fit' 'predict',
    or 'transform' methods. The 'action' function is applied to each found
    algorithm.
    
    action -- a function that is called with as action(class_), where
              class_ is a class that defines the 'fit' or 'predict' method
    """

    # only consider modules and classes once
    if processed_modules is None:
        processed_modules = []
    if processed_classes is None:
        processed_classes = []

    if current_module in processed_modules:
        return
    processed_modules.append(current_module)

    for member_name, member in list(current_module.__dict__.items()):
        if not member_name.startswith('_'):

            # classes
            if (inspect.isclass(member) and
                member not in processed_classes):
                processed_classes.append(member)
                if (hasattr(member, 'fit')
                    or hasattr(member, 'predict')
                    or hasattr(member, 'transform')):
                    action(member)

            # other modules
            elif (inspect.ismodule(member) and
                  member.__name__.startswith(_sklearn_prefix)):
                apply_to_scikits_algorithms(member, action, processed_modules,
                                            processed_classes)
    return processed_classes


_OUTPUTDIM_ERROR = """'output_dim' keyword not supported.

Please set the output dimensionality using sklearn keyword
arguments (e.g., 'n_components', or 'k'). See the docstring of this
class for details."""

def wrap_scikits_classifier(scikits_class):
    """Wrap a sklearn classifier as an MDP Node subclass.

    The wrapper maps these MDP methods to their sklearn equivalents:
    - _stop_training -> fit
    - _label -> predict
    """

    newaxis = mdp.numx.newaxis

    # create a wrapper class for a sklearn classifier
    class ScikitsNode(mdp.ClassifierCumulator):

        def __init__(self, input_dim=None, output_dim=None, dtype=None,
                     **kwargs):

            if output_dim is not None:
                # output_dim and n_components cannot be defined at the same time
                if 'n_components' in kwargs:
                    msg = ("Dimensionality set both by "
                           "output_dim=%d and n_components=%d""")
                    raise ScikitsException(msg % (output_dim,
                                                  kwargs['n_components']))


            super(ScikitsNode, self).__init__(input_dim=input_dim,
                                              output_dim=output_dim,
                                              dtype=dtype)
            self.scikits_alg = scikits_class(**kwargs)

        # ---- re-direct training and execution to the wrapped algorithm

        def _stop_training(self, **kwargs):
            super(ScikitsNode, self)._stop_training(self)
            return self.scikits_alg.fit(self.data, self.labels, **kwargs)

        def _label(self, x):
            return self.scikits_alg.predict(x)[:, newaxis]

        # ---- administrative details

        @staticmethod
        def is_invertible():
            return False

        @staticmethod
        def is_trainable():
            """Return True if the node can be trained, False otherwise."""
            return hasattr(scikits_class, 'fit')

        # NOTE: at this point scikits nodes can only support up to
        # 64-bits floats because some call numpy.linalg.svd, which for
        # some reason does not support higher precisions
        def _get_supported_dtypes(self):
            """Return the list of dtypes supported by this node.
            The types can be specified in any format allowed by numpy.dtype."""
            return ['float32', 'float64']

    # modify class name and docstring
    ScikitsNode.__name__ = scikits_class.__name__ + 'ScikitsLearnNode'
    ScikitsNode.__doc__ = _gen_docstring(scikits_class)

    # change the docstring of the methods to match the ones in sklearn

    # methods_dict maps ScikitsNode method names to sklearn method names
    methods_dict = {'__init__': '__init__',
                    'stop_training': 'fit',
                    'label': 'predict'}
    if hasattr(scikits_class, 'predict_proba'):
        methods_dict['prob'] = 'predict_proba'

    for mdp_name, scikits_name in list(methods_dict.items()):
        mdp_method = getattr(ScikitsNode, mdp_name)
        scikits_method = getattr(scikits_class, scikits_name)
        if hasattr(scikits_method, '__func__'):
            mdp_method.__func__.__doc__ = _gen_docstring(scikits_class,
                                                        scikits_method.__func__)

    if scikits_class.__init__.__doc__ is None:
        try:
            ScikitsNode.__init__.__func__.__doc__ = _gen_docstring(scikits_class)
        except AttributeError:
            # we are in Python3
            ScikitsNode.__init__.__doc__ = _gen_docstring(scikits_class)

    return ScikitsNode


def wrap_scikits_transformer(scikits_class):
    """Wrap a sklearn transformer as an MDP Node subclass.

    The wrapper maps these MDP methods to their sklearn equivalents:
    _stop_training -> fit
    _execute -> transform
    """

    # create a wrapper class for a sklearn transformer
    class ScikitsNode(mdp.Cumulator):

        def __init__(self, input_dim=None, output_dim=None, dtype=None, **kwargs):
            if output_dim is not None:
                raise ScikitsException(_OUTPUTDIM_ERROR)
            super(ScikitsNode, self).__init__(input_dim=input_dim,
                                              output_dim=output_dim,
                                              dtype=dtype)
            self.scikits_alg = scikits_class(**kwargs)

        # ---- re-direct training and execution to the wrapped algorithm

        def _stop_training(self, **kwargs):
            super(ScikitsNode, self)._stop_training(self)
            return self.scikits_alg.fit(self.data, **kwargs)

        def _execute(self, x):
            return self.scikits_alg.transform(x)

        # ---- administrative details

        @staticmethod
        def is_invertible():
            return False

        @staticmethod
        def is_trainable():
            """Return True if the node can be trained, False otherwise."""
            return hasattr(scikits_class, 'fit')

        # NOTE: at this point scikits nodes can only support up to
        # 64-bits floats because some call numpy.linalg.svd, which for
        # some reason does not support higher precisions
        def _get_supported_dtypes(self):
            """Return the list of dtypes supported by this node.
            The types can be specified in any format allowed by numpy.dtype."""
            return ['float32', 'float64']

    # modify class name and docstring
    ScikitsNode.__name__ = scikits_class.__name__ + 'ScikitsLearnNode'
    ScikitsNode.__doc__ = _gen_docstring(scikits_class)

    # change the docstring of the methods to match the ones in sklearn

    # methods_dict maps ScikitsNode method names to sklearn method names
    methods_dict = {'__init__': '__init__',
                    'stop_training': 'fit',
                    'execute': 'transform'}

    for mdp_name, scikits_name in list(methods_dict.items()):
        mdp_method = getattr(ScikitsNode, mdp_name)
        scikits_method = getattr(scikits_class, scikits_name, None)
        if hasattr(scikits_method, '__func__'):
            mdp_method.__func__.__doc__ = _gen_docstring(scikits_class,
                                                        scikits_method.__func__)

    if scikits_class.__init__.__doc__ is None:
        try:
            ScikitsNode.__init__.__func__.__doc__ = _gen_docstring(scikits_class)
        except AttributeError:
            # we are in Python3
            ScikitsNode.__init__.__doc__ = _gen_docstring(scikits_class)

    return ScikitsNode


def wrap_scikits_predictor(scikits_class):
    """Wrap a sklearn transformer as an MDP Node subclass.

    The wrapper maps these MDP methods to their sklearn equivalents:
    _stop_training -> fit
    _execute -> predict
    """

    # create a wrapper class for a sklearn predictor
    class ScikitsNode(mdp.Cumulator):

        def __init__(self, input_dim=None, output_dim=None, dtype=None, **kwargs):
            if output_dim is not None:
                raise ScikitsException(_OUTPUTDIM_ERROR)
            super(ScikitsNode, self).__init__(input_dim=input_dim,
                                              output_dim=output_dim,
                                              dtype=dtype)
            self.scikits_alg = scikits_class(**kwargs)

        # ---- re-direct training and execution to the wrapped algorithm

        def _stop_training(self, **kwargs):
            super(ScikitsNode, self)._stop_training(self)
            return self.scikits_alg.fit(self.data, **kwargs)

        def _execute(self, x):
            return self.scikits_alg.predict(x)

        # ---- administrative details

        @staticmethod
        def is_invertible():
            return False

        @staticmethod
        def is_trainable():
            """Return True if the node can be trained, False otherwise."""
            return hasattr(scikits_class, 'fit')

        # NOTE: at this point scikits nodes can only support up to 64-bits floats
        # because some call numpy.linalg.svd, which for some reason does not
        # support higher precisions
        def _get_supported_dtypes(self):
            """Return the list of dtypes supported by this node.
            The types can be specified in any format allowed by numpy.dtype."""
            return ['float32', 'float64']

    # modify class name and docstring
    ScikitsNode.__name__ = scikits_class.__name__ + 'ScikitsLearnNode'
    ScikitsNode.__doc__ = _gen_docstring(scikits_class)

    # change the docstring of the methods to match the ones in sklearn

    # methods_dict maps ScikitsNode method names to sklearn method names
    methods_dict = {'__init__': '__init__',
                    'stop_training': 'fit',
                    'execute': 'predict'}

    for mdp_name, scikits_name in list(methods_dict.items()):
        mdp_method = getattr(ScikitsNode, mdp_name)
        scikits_method = getattr(scikits_class, scikits_name)
        if hasattr(scikits_method, '__func__'):
            mdp_method.__func__.__doc__ = _gen_docstring(scikits_class,
                                                        scikits_method.__func__)

    if scikits_class.__init__.__doc__ is None:
        try:
            ScikitsNode.__init__.__func__.__doc__ = _gen_docstring(scikits_class)
        except AttributeError:
            # we are in Python3
            ScikitsNode.__init__.__doc__ = _gen_docstring(scikits_class)

    return ScikitsNode


#list candidate nodes
def print_public_members(class_):
    """Print methods of sklearn algorithm.
    """
    print('\n', '-' * 15)
    print('%s (%s)' % (class_.__name__, class_.__module__))
    for attr_name in dir(class_):
        attr = getattr(class_, attr_name)
        #print attr_name, type(attr)
        if not attr_name.startswith('_') and inspect.ismethod(attr):
            print(' -', attr_name)

#apply_to_scikits_algorithms(sklearn, print_public_members)


def wrap_scikits_algorithms(scikits_class, nodes_list):
    """NEED DOCSTRING."""

    name = scikits_class.__name__
    if (name[:4] == 'Base' or name == 'LinearModel'
        or name.startswith('EllipticEnvelop')
        or name.startswith('ForestClassifier')):
        return

    if issubclass(scikits_class, sklearn.base.ClassifierMixin) and \
        hasattr(scikits_class, 'fit'):
        nodes_list.append(wrap_scikits_classifier(scikits_class))
    # Some (abstract) transformers do not implement fit.
    elif hasattr(scikits_class, 'transform') and hasattr(scikits_class, 'fit'):
        nodes_list.append(wrap_scikits_transformer(scikits_class))
    elif hasattr(scikits_class, 'predict') and hasattr(scikits_class, 'fit'):
        nodes_list.append(wrap_scikits_predictor(scikits_class))

scikits_nodes = []
apply_to_scikits_algorithms(sklearn,
                            lambda c: wrap_scikits_algorithms(c, scikits_nodes))

# add scikit nodes to dictionary
#scikits_module = new.module('scikits')
DICT_ = {}
for wrapped_c in scikits_nodes:
    #print wrapped_c.__name__
    DICT_[wrapped_c.__name__] = wrapped_c



# -*- coding: utf-8 -*-

"""These are test functions for MDP classifiers.
"""
from __future__ import division
from builtins import zip
from builtins import range
from past.utils import old_div
from ._tools import *

from mdp import ClassifierNode
from mdp.nodes import (SignumClassifier, PerceptronClassifier,
                       SimpleMarkovClassifier, DiscreteHopfieldClassifier,
                       KMeansClassifier)

def _sigmoid(t):
    return old_div(1.0, (1.0 + numx.exp(-t)))

class _BogusClassifier(ClassifierNode):
    @staticmethod
    def is_trainable():
        return False
    def _label(self, x):
        return [r[0] for r in self.rank(x)]
    def _prob(self, x):
        return [{-1: _sigmoid(sum(xi)), \
                  1: 1 - _sigmoid(sum(xi))} for xi in x]


def testClassifierNode_ranking():
    bc = _BogusClassifier()
    test_data = numx_rand.random((30, 20)) - 0.5
    for r, p in zip(bc.rank(test_data), bc.prob(test_data)):
        # check that the ranking order is correct
        assert p[r[0]] >= p[r[1]], "Rank returns labels in incorrect order"
        # check that the probabilities sum up to 100
        assert 0.999 < p[r[0]] + p[r[1]] < 1.001

def testClassifier_execute_method():
    """Test that the execute result has the correct format when execute_method
    is used.
    """
    bc = _BogusClassifier(execute_method="label")
    data = numx_rand.random((5, 20)) - 0.5
    result = bc.execute(data)
    assert isinstance(result, list)
    assert isinstance(result[0], int)
    bc.execute_method = "prob"
    result = bc.execute(data)
    assert isinstance(result, list)
    assert isinstance(result[0], dict)
    bc.execute_method = "rank"
    result = bc.execute(data)
    assert isinstance(result, list)
    assert isinstance(result[0], list)

def testSignumClassifier():
    c = SignumClassifier()
    res = c.label(mdp.numx.array([[1, 2, -3, -4], [1, 2, 3, 4]]))
    assert c.input_dim == 4
    assert res.tolist() == [-1, 1]

def testPerceptronClassifier():
    or_Classifier = PerceptronClassifier()
    for i in range(100):
        or_Classifier.train(mdp.numx.array([[0., 0.]]), -1)
        or_Classifier.train(mdp.numx.array([[0., 1.], [1., 0.], [1., 1.]]), 1)
    assert or_Classifier.input_dim == 2

    res = or_Classifier.label(mdp.numx.array([[0., 0.], [0., 1.], [1., 0.], [1., 1.]]))
    assert res.tolist() == [-1, 1, 1, 1]

    and_Classifier = PerceptronClassifier()
    for i in range(100):
        and_Classifier.train(mdp.numx.array([[0., 0.], [0., 1.], [1., 0.]]), -1)
        and_Classifier.train(mdp.numx.array([[1., 1.]]), 1)
    res = and_Classifier.label(mdp.numx.array([[0., 0.], [0., 1.], [1., 0.], [1., 1.]]))
    assert res.tolist() == [-1, -1, -1, 1]

    xor_Classifier = PerceptronClassifier()
    for i in range(100):
        xor_Classifier.train(mdp.numx.array([[0., 0.], [1., 1.]]), -1)
        xor_Classifier.train(mdp.numx.array([[0., 1.], [1., 0.]]), 1)
    res = xor_Classifier.label(mdp.numx.array([[0., 0.], [0., 1.], [1., 0.], [1., 1.]]))
    assert res.tolist() != [-1, 1, 1, -1], \
        "Something must be wrong here. XOR is impossible in a single-layered perceptron."


def testSimpleMarkovClassifier():
    mc = SimpleMarkovClassifier(dtype="c")
    text = "after the letter e follows either space or the letters r t or i"

    for word in text.split():
        word = word.lower()

        features = list(zip(" " + word))
        labels = list(word + " ")

        mc.train(mdp.numx.array(features), labels)

    assert mc.input_dim == 1

    num_transitions = 0
    features = mc.features
    for feature, count in list(features.items()):
        if count:
            prob = mc.prob(mdp.numx.array([feature]))
            prob_sum = 0
            for p in prob:
                for k, v in list(p.items()):
                    prob_sum += v
                    if v:
                        num_transitions += 1

            assert abs(prob_sum - 1.0) < 1e-5

    # calculate the number of transitions (the negative set deletes the artefact of two spaces)
    trans = len(set((list(zip("  ".join(text.split()) + " ", \
                         " " + "  ".join(text.split()))))) - set([(' ', ' ')]))
    assert num_transitions == trans

    letters_following_e = [' ', 'r', 't', 'i']
    letters_prob = mc.prob(mdp.numx.array([['e']]))[0]
    prob_sum = 0
    for letter, prob in list(letters_prob.items()):
        prob_sum += prob
        if prob > 1e-5:
            assert letter in letters_following_e

    assert abs(prob_sum - 1.0) < 1e-5

def testDiscreteHopfieldClassifier():
    h = DiscreteHopfieldClassifier()

    memory_size = 100
    patterns = numx.array(
               [numx.sin(numx.linspace(0, 100 * numx.pi, memory_size)) > 0,
                numx.sin(numx.linspace(0, 50 * numx.pi, memory_size)) > 0,
                numx.sin(numx.linspace(0, 20 * numx.pi, memory_size)) > 0,
                numx.sin(numx.linspace(0, 15 * numx.pi, memory_size)) > 0,
                numx.sin(numx.linspace(0, 10 * numx.pi, memory_size)) > 0,
                numx.sin(numx.linspace(0, 5 * numx.pi, memory_size)) > 0,
                numx.sin(numx.linspace(0, 2 * numx.pi, memory_size)) > 0
                ])
    h.train(patterns)
    h.input_dim = memory_size

    for p in patterns:
        # check if patterns are fixpoints
        assert numx.all(p == h.label(numx.array([p])))

    for p in patterns:
        # check, if a noisy pattern is recreated
        noisy = numx.array(p)
        for i in range(len(noisy)):
            if numx.random.random() > 0.95:
                noisy[i] = not noisy[i]
        retrieved = h.label(numx.array([noisy]))
        # Hopfield nets are blind for inversion, need to check either case
        assert numx.all(retrieved == p) or numx.all(retrieved != p)

def testKMeansClassifier():
    num_centroids = 3
    k = KMeansClassifier(num_centroids)
    a = numx.random.rand(50, 2)
    k.train(a)
    res = k.label(a)

    # check that the number of centroids is correct
    assert len(set(res)) == num_centroids

    k = KMeansClassifier(2)
    a1 = numx.random.rand(50, 2) - 1
    a2 = numx.random.rand(50, 2) + 1
    k.train(a1)
    k.train(a2)
    res1 = k.label(a1)
    res2 = k.label(a2)
    # check that both clusters are completely identified and different
    assert (len(set(res1)) == 1 and
            len(set(res2)) == 1 and
            set(res1) != set(res2)
            ), ("Error in K-Means classifier. "
                "This might be a bug or just a local minimum.")



"""These are test functions for MDP contributed nodes.
"""
from __future__ import division
from builtins import str
from builtins import range
from past.utils import old_div

from ._tools import *
from .test_ICANode import verify_ICANode, verify_ICANodeMatrices

requires_joblib = skip_on_condition(
    "not mdp.config.has_caching",
    "This test requires the 'joblib' module.")

def _s_shape(theta):
    """
    returns x,y
      a 2-dimensional S-shaped function
      for theta ranging from 0 to 1
    """
    t = 3*numx.pi * (theta-0.5)
    x = numx.sin(t)
    y = numx.sign(t)*(numx.cos(t)-1)
    return x,y

def _s_shape_1D(n):
    t = numx.linspace(0., 1., n)
    x, z = _s_shape(t)
    y = numx.linspace(0., 5., n)
    return x, y, z, t

def _s_shape_2D(nt, ny):
    t, y = numx.meshgrid(numx.linspace(0., 1., nt),
                         numx.linspace(0., 2., ny))
    t = t.flatten()
    y = y.flatten()
    x, z = _s_shape(t)
    return x, y, z, t

def _compare_neighbors(orig, proj, k):
    n = orig.shape[0]
    err = numx.zeros((n,))
    # compare neighbors indices
    for i in range(n):
        # neighbors in original space
        dist = orig - orig[i,:]
        orig_nbrs = numx.argsort((dist**2).sum(1))[1:k+1]
        orig_nbrs.sort()
        # neighbors in projected space
        dist = proj - proj[i,:]
        proj_nbrs = numx.argsort((dist**2).sum(1))[1:k+1]
        proj_nbrs.sort()
        for idx in orig_nbrs:
            if idx not in proj_nbrs:
                err[i] += 1
    return err

def test_JADENode():
    trials = 3
    for i in range(trials):
        try:
            ica = mdp.nodes.JADENode(limit = 10**(-decimal))
            ica2 = ica.copy()
            verify_ICANode(ica, rand_func=numx_rand.exponential)
            verify_ICANodeMatrices(ica2)
            return
        except Exception:
            if i == trials - 1:
                raise

def test_NIPALSNode():
    line_x = numx.zeros((1000,2),"d")
    line_y = numx.zeros((1000,2),"d")
    line_x[:,0] = numx.linspace(-1,1,num=1000,endpoint=1)
    line_y[:,1] = numx.linspace(-0.2,0.2,num=1000,endpoint=1)
    mat = numx.concatenate((line_x,line_y))
    des_var = std(mat,axis=0)
    utils.rotate(mat,uniform()*2*numx.pi)
    mat += uniform(2)
    pca = mdp.nodes.NIPALSNode(conv=1E-15, max_it=1000)
    pca.train(mat)
    act_mat = pca.execute(mat)
    assert_array_almost_equal(mean(act_mat,axis=0),\
                              [0,0],decimal)
    assert_array_almost_equal(std(act_mat,axis=0),\
                              des_var,decimal)
    # test a bug in v.1.1.1, should not crash
    pca.inverse(act_mat[:,:1])
    # try standard PCA on the same data and compare the eigenvalues
    pca2 = mdp.nodes.PCANode()
    pca2.train(mat)
    pca2.stop_training()
    assert_array_almost_equal(pca2.d, pca.d, decimal)

def test_NIPALSNode_desired_variance():
    mat, mix, inp = get_random_mix(mat_dim=(1000, 3))
    # first make them white
    pca = mdp.nodes.WhiteningNode()
    pca.train(mat)
    mat = pca.execute(mat)
    # set the variances
    mat *= [0.6,0.3,0.1]
    #mat -= mat.mean(axis=0)
    pca = mdp.nodes.NIPALSNode(output_dim=0.8)
    pca.train(mat)
    out = pca.execute(mat)
    # check that we got exactly two output_dim:
    assert pca.output_dim == 2
    assert out.shape[1] == 2
    # check that explained variance is > 0.8 and < 1
    assert (pca.explained_variance > 0.8 and pca.explained_variance < 1)

def test_LLENode():
    # 1D S-shape in 3D
    n, k = 50, 2
    x, y, z, t = _s_shape_1D(n)
    data = numx.asarray([x,y,z]).T

    res = mdp.nodes.LLENode(k, output_dim=1, svd=False)(data)
    # check that the neighbors are the same
    err = _compare_neighbors(data, res, k)
    assert err.max() == 0

    # with svd=True
    res = mdp.nodes.LLENode(k, output_dim=1, svd=True)(data)
    err = _compare_neighbors(data, res, k)
    assert err.max() == 0
    return

    #TODO: fix this test!
    # 2D S-shape in 3D
    nt, ny = 40, 15
    n, k = nt*ny, 8
    x, y, z, t = _s_shape_2D(nt, ny)
    data = numx.asarray([x,y,z]).T
    res = mdp.nodes.LLENode(k, output_dim=2, svd=True)(data)
    res[:,0] /= res[:,0].std()
    res[:,1] /= res[:,1].std()

    # test alignment
    yval = y[::nt]
    tval = t[:ny]
    for yv in yval:
        idx = numx.nonzero(y==yv)[0]
        err = abs(res[idx,1]-res[idx[0],1]).max()
        assert err<0.01,\
               'Projection should be aligned as original space: %s'%(str(err))
    for tv in tval:
        idx = numx.nonzero(t==tv)[0]
        err = abs(res[idx,0]-res[idx[0],0]).max()
        assert err<0.01,\
               'Projection should be aligned as original space: %s'%(str(err))

def test_LLENode_outputdim_float_bug():
    # 1D S-shape in 3D, output_dim
    n, k = 50, 2
    x, y, z, t = _s_shape_1D(n)
    data = numx.asarray([x,y,z]).T

    res = mdp.nodes.LLENode(k, output_dim=0.9, svd=True)(data)
    # check that the neighbors are the same
    err = _compare_neighbors(data, res, k)
    assert err.max() == 0

def test_HLLENode():
    # 1D S-shape in 3D
    n, k = 250, 4
    x, y, z, t = _s_shape_1D(n)
    data = numx.asarray([x,y,z]).T

    res = mdp.nodes.HLLENode(k, r=0.001, output_dim=1, svd=False)(data)
    # check that the neighbors are the same
    err = _compare_neighbors(data, res, k)
    assert err.max() == 0

    # with svd=True
    res = mdp.nodes.HLLENode(k, r=0.001, output_dim=1, svd=True)(data)
    err = _compare_neighbors(data, res, k)
    assert err.max() == 0

    # 2D S-shape in 3D
    nt, ny = 40, 15
    n, k = nt*ny, 8
    x, y, z, t = _s_shape_2D(nt, ny)
    data = numx.asarray([x,y,z]).T
    res = mdp.nodes.HLLENode(k, r=0.001, output_dim=2, svd=False)(data)
    res[:,0] /= res[:,0].std()
    res[:,1] /= res[:,1].std()

    # test alignment
    yval = y[::nt]
    tval = t[:ny]
    for yv in yval:
        idx = numx.nonzero(y==yv)[0]
        assert numx.all(res[idx,1]-res[idx[0],1]<1e-2),\
               'Projection should be aligned as original space'
    for tv in tval:
        idx = numx.nonzero(t==tv)[0]
        assert numx.all(res[idx,0]-res[idx[0],0]<1e-2),\
               'Projection should be aligned as original space'

def test_XSFANode():
    T = 5000
    N = 3
    src = numx_rand.random((T, N))*2-1
    # create three souces with different speeds
    fsrc = numx_fft.rfft(src, axis=0)

    for i in range(N):
        fsrc[(i+1)*(old_div(T,10)):, i] = 0.

    src = numx_fft.irfft(fsrc,axis=0)
    src -= src.mean(axis=0)
    src /= src.std(axis=0)

    #mix = sigmoid(numx.dot(src, mdp.utils.random_rot(3)))
    mix = src

    flow = mdp.Flow([mdp.nodes.XSFANode()])
    # let's test also chunk-mode training
    flow.train([[mix[:old_div(T,2), :], mix[old_div(T,2):, :]]])

    out = flow(mix)
    #import bimdp
    #tr_filename = bimdp.show_training(flow=flow,
    #                                  data_iterators=[[mix[:T/2, :], mix[T/2:, :]]])
    #ex_filename, out = bimdp.show_execution(flow, x=mix)

    corrs = mdp.utils.cov_maxima(mdp.utils.cov2(out, src))
    assert min(corrs) > 0.8, ('source/estimate minimal'
                              ' covariance: %g' % min(corrs))




from future import standard_library
standard_library.install_aliases()
from builtins import zip
from builtins import range
"""These are test functions for hinet.
"""

import py.test
import io
import mdp.hinet as mh
from ._tools import *


_get_new_flow = lambda: mdp.Flow([
        mdp.nodes.NoiseNode(),
        mdp.nodes.SFANode()])

_get_new_nodes = lambda: [
        mdp.nodes.CuBICANode(input_dim=1, whitened=True),
        mdp.nodes.CuBICANode(input_dim=2, whitened=True),
        mdp.nodes.CuBICANode(input_dim=1, whitened=True),
        ]

_get_sigle_node = lambda: mdp.nodes.CuBICANode(input_dim=2, whitened=True)

NODES = [(mh.FlowNode, [_get_new_flow], None),
         (mh.Layer, [_get_new_nodes], None),
         (mh.CloneLayer, [_get_sigle_node, 2], None),
         ]

def test_FlowNode_training():
    flow = mdp.Flow([mdp.nodes.PolynomialExpansionNode(degree=2),
                     mdp.nodes.PCANode(output_dim=15, reduce=True),
                     mdp.nodes.PolynomialExpansionNode(degree=2),
                     mdp.nodes.PCANode(output_dim=3, reduce=True)])
    flownode = mh.FlowNode(flow)
    x = numx_rand.random([300,20])
    while flownode.get_remaining_train_phase() > 0:
        flownode.train(x)
        flownode.stop_training()
    flownode.execute(x)

def test_FlowNode_trainability():
    flow = mdp.Flow([mdp.nodes.PolynomialExpansionNode(degree=2)])
    flownode = mh.FlowNode(flow)
    assert flownode.is_trainable() is False
    flow = mdp.Flow([mdp.nodes.PolynomialExpansionNode(degree=2),
                     mdp.nodes.PCANode(output_dim=15),
                     mdp.nodes.PolynomialExpansionNode(degree=2),
                     mdp.nodes.PCANode(output_dim=3)])
    flownode = mh.FlowNode(flow)
    assert flownode.is_trainable() is True

def test_FlowNode_invertibility():
    flow = mdp.Flow([mdp.nodes.PolynomialExpansionNode(degree=2)])
    flownode = mh.FlowNode(flow)
    assert flownode.is_invertible() is False
    flow = mdp.Flow([mdp.nodes.PCANode(output_dim=15),
                     mdp.nodes.SFANode(),
                     mdp.nodes.PCANode(output_dim=3)])
    flownode = mh.FlowNode(flow)
    assert flownode.is_invertible() is True

def test_FlowNode_pretrained_node():
    x = numx_rand.random([100,10])
    pretrained_node = mdp.nodes.PCANode(output_dim=6)
    pretrained_node.train(x)
    pretrained_node.stop_training()
    flow = mdp.Flow([pretrained_node,
                     mdp.nodes.PolynomialExpansionNode(degree=2),
                     mdp.nodes.PCANode(output_dim=3)])
    flownode = mh.FlowNode(flow)
    while flownode.get_remaining_train_phase() > 0:
        flownode.train(x)
        flownode.stop_training()
    flownode.execute(x)

def test_FlowNode_fix_nodes_dimensions1():
    x = numx_rand.random([100,10])
    last_node = mdp.nodes.IdentityNode()
    flow = mdp.Flow([mdp.nodes.PCANode(output_dim=3),
                     mdp.nodes.IdentityNode(),
                     last_node])
    flownode = mh.FlowNode(flow)
    flownode.train(x)
    flownode.stop_training()
    # check that the dimensions of NoiseNode and FlowNode where all set
    # by calling _fix_nodes_dimensions
    assert flownode.output_dim == 3
    assert last_node.input_dim == 3
    assert last_node.output_dim == 3

def test_FlowNode_fix_nodes_dimensions2():
    flow = mdp.Flow([mdp.nodes.IdentityNode(),
                     mdp.nodes.IdentityNode()])
    flownode = mh.FlowNode(flow)
    # this should fail, since the internal nodes don't have fixed dims
    py.test.raises(mdp.InconsistentDimException,
                   lambda: flownode.set_output_dim(10))
    x = numx_rand.random([100,10])
    flownode.execute(x)
    assert flownode.output_dim == 10

def test_FlowNode_fix_nodes_dimensions3():
    flow = mdp.Flow([mdp.nodes.IdentityNode()])
    flownode = mh.FlowNode(flow)
    # for a single node this should not raise an Exception
    flownode.set_output_dim(10)
    x = numx_rand.random([100,10])
    flownode.execute(x)

def test_FlowNode_pretrained_flow():
    flow = mdp.Flow([mdp.nodes.PolynomialExpansionNode(degree=2),
                     mdp.nodes.PCANode(output_dim=15, reduce=True),
                     mdp.nodes.PolynomialExpansionNode(degree=2),
                     mdp.nodes.PCANode(output_dim=3, reduce=True)])
    flownode = mh.FlowNode(flow)
    x = numx_rand.random([300,20])
    while flownode.get_remaining_train_phase() > 0:
        flownode.train(x)
        flownode.stop_training()
    # build new flownode with the trained nodes
    flownode = mh.FlowNode(flow)
    assert not flownode.is_training()
    flownode.execute(x)

def test_FlowNode_copy1():
    flow = mdp.Flow([mdp.nodes.PCANode(), mdp.nodes.SFANode()])
    flownode = mh.FlowNode(flow)
    flownode.copy()

def test_FlowNode_copy2():
    # Test that the FlowNode copy method delegates to internal nodes.
    class CopyFailException(Exception):
        pass
    class CopyFailNode(mdp.Node):
        def copy(self, protocol=None):
            raise CopyFailException()
    flow = mdp.Flow([mdp.Node(), CopyFailNode()])
    flownode = mh.FlowNode(flow)
    py.test.raises(CopyFailException, flownode.copy)

def _pca_nodes(input_dims, output_dims):
    return [mdp.nodes.PCANode(input_dim=ind, output_dim=outd)
            for ind, outd in zip(input_dims, output_dims)]

def test_Layer():
    layer = mh.Layer(_pca_nodes([10, 17, 3], [5, 3, 1]))
    x = numx_rand.random([100,30]).astype('f')
    layer.train(x)
    y = layer.execute(x)
    assert layer.dtype == numx.dtype('f')
    assert y.dtype == layer.dtype

def test_Layer_invertibility():
    layer = mh.Layer(_pca_nodes([10, 17, 3], [10, 17, 3]))
    x = numx_rand.random([100,30]).astype('f')
    layer.train(x)
    y = layer.execute(x)
    x_inverse = layer.inverse(y)
    assert numx.all(numx.absolute(x - x_inverse) < 0.001)

def test_Layer_invertibility2():
    # reduce the dimensions, so input_dim != output_dim
    layer = mh.Layer(_pca_nodes([10, 17, 3], [8, 12, 3]))
    x = numx_rand.random([100,30]).astype('f')
    layer.train(x)
    y = layer.execute(x)
    layer.inverse(y)

def test_SameInputLayer():
    layer = mh.SameInputLayer(_pca_nodes([10, 10, 10], [5, 3, 1]))
    x = numx_rand.random([100,10]).astype('f')
    layer.train(x)
    y = layer.execute(x)
    assert layer.dtype == numx.dtype('f')
    assert y.dtype == layer.dtype

def test_CloneLayer():
    node = mdp.nodes.PCANode(input_dim=10, output_dim=5)
    x = numx_rand.random([10,70]).astype('f')
    layer = mh.CloneLayer(node, 7)
    layer.train(x)
    y = layer.execute(x)
    assert layer.dtype == numx.dtype('f')
    assert y.dtype == layer.dtype

def test_SwitchboardInverse1():
    sboard = mh.Switchboard(input_dim=3,
                            connections=[2,0,1])
    assert sboard.is_invertible()
    y = numx.array([[2,3,4],[5,6,7]])
    x = sboard.inverse(y)
    assert numx.all(x == numx.array([[3,4,2],[6,7,5]]))

def testSwitchboardInverse2():
    sboard = mh.Switchboard(input_dim=3,
                            connections=[2,1,1])
    assert not sboard.is_invertible()

## Tests for MeanInverseSwitchboard ##

def test_MeanInverseSwitchboard1():
    sboard = mh.MeanInverseSwitchboard(input_dim=3,
                                       connections=[0,0,2])
    assert sboard.is_invertible()
    y = numx.array([[2,4,3],[1,1,7]])
    x = sboard.inverse(y)
    assert numx.all(x == numx.array([[3,0,3],[1,0,7]]))

def test_MeanInverseSwitchboard2():
    sboard = mh.MeanInverseSwitchboard(input_dim=3,
                                       connections=[1,1,1,2,2])
    assert sboard.is_invertible()
    y = numx.array([[2,4,0,1,1],[3,3,3,2,4]])
    x = sboard.inverse(y)
    assert numx.all(x == numx.array([[0,2,1],[0,3,3]]))

## Tests for ChannelSwitchboard ##

def testOutChannelInput():
    sboard = mh.ChannelSwitchboard(input_dim=6,
                                   connections=[5,5,
                                                0,1],
                                   out_channel_dim=2,
                                   in_channel_dim=2)
    assert numx.all(sboard.get_out_channel_input(0) ==
                    numx.array([5,5]))
    assert numx.all(sboard.get_out_channel_input(1) ==
                    numx.array([0,1]))

def testOutChannelsInputChannels():
    sboard = mh.ChannelSwitchboard(input_dim=6,
                                   connections=[5,5, # out chan 1
                                                0,1], # out chan 2
                                   out_channel_dim=2,
                                   in_channel_dim=2)
    # note that there are 3 input channels
    assert numx.all(sboard.get_out_channels_input_channels(0) ==
                    numx.array([2]))
    assert numx.all(sboard.get_out_channels_input_channels(1) ==
                    numx.array([0]))
    assert numx.all(sboard.get_out_channels_input_channels([0,1]) ==
                    numx.array([0,2]))

## Tests for Rectangular2dSwitchboard ##

def testRect2dRouting1():
    sboard = mh.Rectangular2dSwitchboard(in_channels_xy=(3,2),
                                         in_channel_dim=2,
                                         field_channels_xy=(2,1),
                                         field_spacing_xy=1)
    assert numx.all(sboard.connections ==
                           numx.array([0, 1, 2, 3, 2, 3, 4, 5, 6, 7,
                                       8, 9, 8, 9, 10, 11]))
    x = numx.array([list(range(0, sboard.input_dim)),
                    list(range(101, 101+sboard.input_dim))])
    sboard.execute(x)
    # test generated switchboard
    channel_sboard = sboard.get_out_channel_node(0)
    channel_sboard.execute(x)

def testRect2dRouting2():
    sboard = mh.Rectangular2dSwitchboard(in_channels_xy=(2,4),
                                         in_channel_dim=1,
                                         field_channels_xy=(1,2),
                                         field_spacing_xy=(1,2))
    assert numx.all(sboard.connections ==
                    numx.array([0, 2, 1, 3, 4, 6, 5, 7]))
    x = numx.array([list(range(0, sboard.input_dim)),
                    list(range(101, 101+sboard.input_dim))])
    sboard.execute(x)
    # test generated switchboard
    channel_sboard = sboard.get_out_channel_node(0)
    channel_sboard.execute(x)

def testRect2dRouting3():
    sboard = mh.Rectangular2dSwitchboard(in_channels_xy=(2,4),
                                         in_channel_dim=1,
                                         field_channels_xy=2,
                                         field_spacing_xy=(1,2))
    assert (sboard.connections ==
            numx.array([0, 1, 2, 3, 4, 5, 6, 7])).all()

def testRect2dRouting4():
    sboard = mh.Rectangular2dSwitchboard(in_channels_xy=4,
                                         in_channel_dim=1,
                                         field_channels_xy=(3,2),
                                         field_spacing_xy=(1,2))
    assert (sboard.connections ==
            numx.array([0, 1, 2, 4, 5, 6,
                        1, 2, 3, 5, 6, 7,
                        8, 9, 10, 12, 13, 14,
                        9, 10, 11, 13, 14, 15])).all()

def testRect2d_get_out_channel_node():
    sboard = mh.Rectangular2dSwitchboard(in_channels_xy=(5,4),
                                         in_channel_dim=2,
                                         field_channels_xy=(3,2),
                                         field_spacing_xy=(1,2))
    x = numx.array([list(range(0, sboard.input_dim)),
                    list(range(101, 101+sboard.input_dim))])
    y = sboard.execute(x)
    # routing layer
    nodes = [sboard.get_out_channel_node(index)
             for index in range(sboard.output_channels)]
    layer = mh.SameInputLayer(nodes)
    layer_y = layer.execute(x)
    assert (y == layer_y).all()

def test_Rect2d_exception_1():
    bad_args = dict(in_channels_xy=(12,8),
                    # 3 is the problematic value:
                    field_channels_xy=(4,3),
                    field_spacing_xy=2,
                    in_channel_dim=3,
                    ignore_cover=False)
    with py.test.raises(mh.Rectangular2dSwitchboardException):
        mh.Rectangular2dSwitchboard(**bad_args)

def test_Rect2d_exception_2():
    bad_args = dict(in_channels_xy=(12,8),
                    # 9 is the problematic value:
                    field_channels_xy=(4,9),
                    field_spacing_xy=2,
                    in_channel_dim=3,
                    ignore_cover=False)
    with py.test.raises(mh.Rectangular2dSwitchboardException):
        mh.Rectangular2dSwitchboard(**bad_args)

def test_Rect2d_exception_3():
    bad_args = dict(in_channels_xy=(12,8),
                    # 9 is the problematic value:
                    field_channels_xy=(4,9),
                    field_spacing_xy=2,
                    in_channel_dim=3,
                    ignore_cover=True)
    with py.test.raises(mh.Rectangular2dSwitchboardException):
        mh.Rectangular2dSwitchboard(**bad_args)

## Tests for DoubleRect2dSwitchboard ##

def test_Rect_double_routing_1():
    sboard = mh.DoubleRect2dSwitchboard(in_channels_xy=4,
                                        field_channels_xy=2,
                                        in_channel_dim=1)
    assert (sboard.connections ==
            numx.array([0,1,4,5, 2,3,6,7, 8,9,12,13, 10,11,14,15,
                        # uneven fields
                        5,6,9,10])).all()
    x = numx.array([list(range(0, sboard.input_dim)),
                    list(range(101, 101+sboard.input_dim))])
    sboard.execute(x)

def test_Rect_double_routing_2():
    sboard = mh.DoubleRect2dSwitchboard(in_channels_xy=(6,4),
                                        field_channels_xy=(2,2),
                                        in_channel_dim=1)
    assert (sboard.connections ==
            numx.array([0,1,6,7, 2,3,8,9, 4,5,10,11, 12,13,18,19,
                        14,15,20,21, 16,17,22,23,
                        # uneven fields
                        7,8,13,14, 9,10,15,16])).all()
    x = numx.array([list(range(0, sboard.input_dim)),
                    list(range(101, 101+sboard.input_dim))])
    sboard.execute(x)

def test_Rect_double_routing_3():
    sboard = mh.DoubleRect2dSwitchboard(in_channels_xy=(4,6),
                                        field_channels_xy=2,
                                        in_channel_dim=1)
    assert (sboard.connections ==
            numx.array([0,1,4,5, 2,3,6,7, 8,9,12,13, 10,11,14,15,
                        16,17,20,21, 18,19,22,23,
                        # uneven fields
                        5,6,9,10, 13,14,17,18])).all()
    x = numx.array([list(range(0, sboard.input_dim)),
                    list(range(101, 101+sboard.input_dim))])
    sboard.execute(x)

## Tests for DoubleRhomb2dSwitchboard ##

def test_DoubleRhomb_routing_1():
    sboard = mh.DoubleRhomb2dSwitchboard(long_in_channels_xy=(3,2),
                                         diag_field_channels=2,
                                         in_channel_dim=1)
    assert (sboard.connections ==
            numx.array([1,6,7,4])).all()
    x = numx.array([list(range(0, sboard.input_dim)),
                    list(range(101, 101+sboard.input_dim))])
    sboard.execute(x)

def test_DoubleRhomd_routing_2():
    sboard = mh.DoubleRhomb2dSwitchboard(long_in_channels_xy=(2,3),
                                         diag_field_channels=2,
                                         in_channel_dim=1)
    assert (sboard.connections ==
            numx.array([6,2,3,7])).all()
    x = numx.array([list(range(0, sboard.input_dim)),
                    list(range(101, 101+sboard.input_dim))])
    sboard.execute(x)

def test_DoubleRhomd_routing_3():
    sboard = mh.DoubleRhomb2dSwitchboard(long_in_channels_xy=(4,2),
                                         diag_field_channels=2,
                                         in_channel_dim=1)
    assert (sboard.connections ==
            numx.array([1,8,9,5, 2,9,10,6])).all()
    x = numx.array([list(range(0, sboard.input_dim)),
                    list(range(101, 101+sboard.input_dim))])
    sboard.execute(x)

def test_DoubleRhomd_routing_4():
    sboard = mh.DoubleRhomb2dSwitchboard(long_in_channels_xy=(2,4),
                                         diag_field_channels=2,
                                         in_channel_dim=1)
    assert (sboard.connections ==
            numx.array([8,2,3,9, 9,4,5,10])).all()
    x = numx.array([list(range(0, sboard.input_dim)),
                    list(range(101, 101+sboard.input_dim))])
    sboard.execute(x)

def test_DoubleRhomd_routing_5():
    sboard = mh.DoubleRhomb2dSwitchboard(long_in_channels_xy=4,
                                         diag_field_channels=2,
                                         in_channel_dim=1)
    assert (sboard.connections ==
            numx.array([1,16,17,5,
                        2,17,18,6,
                        5,19,20,9,
                        6,20,21,10,
                        9,22,23,13,
                        10,23,24,14])).all()
    x = numx.array([list(range(0, sboard.input_dim)),
                    list(range(101, 101+sboard.input_dim))])
    sboard.execute(x)

def test_DoubleRhomd_routing_6():
    sboard = mh.DoubleRhomb2dSwitchboard(long_in_channels_xy=(7,4),
                                         diag_field_channels=4,
                                         in_channel_dim=1)
    x = numx.array([list(range(0, sboard.input_dim)),
                    list(range(101, 101+sboard.input_dim))])
    sboard.execute(x)

def test_DoubleRhomd_routing_7():
    sboard = mh.DoubleRhomb2dSwitchboard(long_in_channels_xy=(4,7),
                                         diag_field_channels=4,
                                         in_channel_dim=1)
    x = numx.array([list(range(0, sboard.input_dim)),
                    list(range(101, 101+sboard.input_dim))])
    sboard.execute(x)

def test_DoubleRhomd_routing_8():
    sboard = mh.DoubleRhomb2dSwitchboard(long_in_channels_xy=(6,7),
                                         diag_field_channels=4,
                                         in_channel_dim=1)
    x = numx.array([list(range(0, sboard.input_dim)),
                    list(range(101, 101+sboard.input_dim))])
    sboard.execute(x)

def test_hinet_simple_net():
    switchboard = mh.Rectangular2dSwitchboard(in_channels_xy=(12,8),
                                              field_channels_xy=4,
                                              field_spacing_xy=2,
                                              in_channel_dim=3)

    node = mdp.nodes.PCANode(input_dim=4*4*3, output_dim=5)
    flownode = mh.FlowNode(mdp.Flow([node,]))
    layer = mh.CloneLayer(flownode, switchboard.output_channels)
    flow = mdp.Flow([switchboard, layer])
    x = numx_rand.random([5, switchboard.input_dim])
    flow.train(x)

def pytest_funcarg__noisenode(request):
    return mdp.nodes.NoiseNode(input_dim=20*20,
                               noise_args=(0, 0.0001))

def test_SFA_net(noisenode):
    sfa_node = mdp.nodes.SFANode(input_dim=20*20, output_dim=10, dtype='f')
    switchboard = mh.Rectangular2dSwitchboard(in_channels_xy=100,
                                              field_channels_xy=20,
                                              field_spacing_xy=10)
    flownode = mh.FlowNode(mdp.Flow([noisenode, sfa_node]))
    sfa_layer = mh.CloneLayer(flownode, switchboard.output_channels)
    flow = mdp.Flow([switchboard, sfa_layer])
    train_gen = numx.cast['f'](numx_rand.random((3, 10, 100*100)))
    flow.train([None, train_gen])

def testHiNetHTML(noisenode):
    # create some flow for testing
    sfa_node = mdp.nodes.SFANode(input_dim=20*20, output_dim=10)
    switchboard = mh.Rectangular2dSwitchboard(in_channels_xy=100,
                                              field_channels_xy=20,
                                              field_spacing_xy=10)
    flownode = mh.FlowNode(mdp.Flow([noisenode, sfa_node]))
    sfa_layer = mh.CloneLayer(flownode, switchboard.output_channels)
    flow = mdp.Flow([switchboard, sfa_layer])
    # create dummy file to write the HTML representation
    html_file = io.StringIO()
    hinet_html = mdp.hinet.HiNetHTMLVisitor(html_file)
    hinet_html.convert_flow(flow)
    html_file.close()

def testHiNetXHTML():
    # create some flow for testing
    sfa_node = mdp.nodes.SFANode(input_dim=20*20, output_dim=10)
    flow = mdp.Flow([sfa_node])
    # create dummy file to write the HTML representation
    html_file = io.StringIO()
    hinet_html = mdp.hinet.HiNetXHTMLVisitor(html_file)
    hinet_html.convert_flow(flow)
    html_file.close()



from builtins import range
import mdp.parallel as parallel
from ._tools import *

def test_ParallelGaussianClassifier():
    """Test ParallelGaussianClassifier."""
    precision = 6
    xs = [numx_rand.random([4,5]) for _ in range(8)]
    labels = [1,2,1,1,2,3,2,3]
    node = mdp.nodes.GaussianClassifier()
    pnode = parallel.ParallelGaussianClassifier()
    for i, x in enumerate(xs):
        node.train(x, labels[i])
    node.stop_training()
    pnode1 = pnode.fork()
    pnode2 = pnode.fork()
    for i, x in enumerate(xs):
        if i % 2:
            pnode1.train(x, labels[i])
        else:
            pnode2.train(x, labels[i])
    pnode.join(pnode1)
    pnode.join(pnode2)
    pnode.stop_training()
    # check that results are the same for all object classes
    for i in range(3):
        assert_array_almost_equal(node.inv_covs[i], pnode.inv_covs[i],
                                  precision)
        assert_array_almost_equal(node.means[i], pnode.means[i],
                                  precision)
        assert node.p[i] == pnode.p[i]
        assert node.p[i] == pnode.p[i]
        assert node.labels[i] == pnode.labels[i]
        

def test_ParallelNearestMeanClassifier():
    """Test ParallelGaussianClassifier."""
    precision = 6
    xs = [numx_rand.random([4,5]) for _ in range(8)]
    labels = [1,2,1,1,2,3,2,3]
    node = mdp.nodes.NearestMeanClassifier()
    pnode = parallel.ParallelNearestMeanClassifier()
    for i, x in enumerate(xs):
        node.train(x, labels[i])
    node.stop_training()
    pnode1 = pnode.fork()
    pnode2 = pnode.fork()
    for i, x in enumerate(xs):
        if i % 2:
            pnode1.train(x, labels[i])
        else:
            pnode2.train(x, labels[i])
    pnode.join(pnode1)
    pnode.join(pnode2)
    pnode.stop_training()
    # check that results are the same for all object classes
    assert_array_almost_equal(node.ordered_means, pnode.ordered_means,
                              precision)
    for key in node.label_means:
        assert_array_almost_equal(node.label_means[key],
                                  pnode.label_means[key], precision)
        assert node.n_label_samples[key] == pnode.n_label_samples[key]
        

def test_ParallelKNNClassifier():
    """Test ParallelGaussianClassifier."""
    precision = 6
    xs = [numx_rand.random([3,2]) for _ in range(8)]
    labels = [1,2,1,1,2,3,2,3]
    node = mdp.nodes.KNNClassifier()
    pnode = parallel.ParallelKNNClassifier()
    for i, x in enumerate(xs):
        node.train(x, labels[i])
    node.stop_training()
    pnode1 = pnode.fork()
    pnode2 = pnode.fork()
    for i, x in enumerate(xs):
        if i < 4:
            pnode1.train(x, labels[i])
        else:
            pnode2.train(x, labels[i])
    pnode.join(pnode1)
    pnode.join(pnode2)
    pnode.stop_training()
    # check that results are the same for all object classes
    assert_array_almost_equal(node.samples, pnode.samples,
                              precision)
    assert node.n_samples == pnode.n_samples



"""
The BiMDP package is an extension of the pure feed-forward flow concept in MDP.

It defines a framework for far more general flow sequences, involving
top-down processes (e.g. for error backpropagation) or even loops.
So the 'bi' in BiMDP primarily stands for 'bidirectional'.

BiMDP is implemented by extending both the Node and the Flow concept. Both the
new BiNode and BiFlow classes are downward compatible with the classical
Nodes and Flows, allowing them to be combined with BiMDP elements.

The first fundamental addition in BiMDP is that BiNodes can specify a target
node for their output, to continue the flow execution at the specified target
node. The second new feature is that Nodes can use messages to propagate
arbitrary information, in addition to the standard single array data.
A BiFlow is needed to enable these features, and the BiNode class has adds
convenience functionality to help with this.

Another important addition are the inspection capapbilities (e.g.,
bimdo.show_training), which create and interactive HTML representation of the
data flow. This makes debugging much easier and can also be extended to
visualize data (see the demos in the test folder).

BiMDP fully supports and extends the HiNet and the Parallel packages.


New BiMDP concepts: Jumps and Messages
======================================

Jump targets are numbers (relative position in the flow) or strings, which are
then compared to the optional node_id. The target number 0 refers to the node
itself.
During execution a node can also use the value of EXIT_TARGET (which is
currently just 'exit') as target value to end the execution. The BiFlow
will then return the last output as result.

Messages are standard Python dictionaries to transport information
that would not fit well into the standard x array. The dict keys also support
target specifications and other magic for more convenient usage.
This is described in more detail in the BiNode module.
"""

### T O D O ###

# TODO: add a target seperator that does not remove the key. Could use
#    -> remove key
#    --> remove one '-' on entry
#    => do not remove the key
#  Note that adding this kind of magic is relatively cheap in BiNode,
#  in parsing first check just for > .

# TODO: add wildcard support for node_id in message keys.
#    Simply tread the node_id part of the key as a regex and check for match.
#    This adds an overhead of about 1 sec per 100,000 messages.

# TODO: Terminate execution if both x and msg are None? This could help in
#    the stop_training execution, but could lead to strange results
#    during normal execution.
#    We could add a check before the result is returned in execute.

# TODO: support dictionary methods like 'keys' in BiFlow? 

# TODO: add workaround for Google Chrome issue once a solution for
#    http://code.google.com/p/chromium/issues/detail?id=47416
#    is in place.

# TODO: Implement more internal checks for node output result?
#    Check that last element is not None?

# TODO: implement switchlayer, a layer where each column represents a different
#    target, so the target value determines which nodes are used

# TODO: show more information in trace slides via mouse hover,
#    or enable some kind of folding (might be possible via CSS like suckerfish)


from .binode import (
    BiNodeException, BiNode, PreserveDimBiNode, MSG_ID_SEP, binode_coroutine
)
from .biclassifier import BiClassifier
from .biflow import (
    MessageResultContainer, BiFlowException, BiFlow, BiCheckpointFlow,
    EXIT_TARGET
)
# the inspection stuff is considered a core functionality
from .inspection import *

from .test import test

from . import nodes
from . import hinet
from . import parallel

del binode
del biflow
del inspection

from mdp.utils import fixup_namespace
fixup_namespace(__name__, None,
                ('binode',
                 'biclassifier',
                 'biflow',
                 'inspection',
                 ))
del fixup_namespace



"""Tools for the test- and benchmark functions."""
from __future__ import division
from builtins import next
from builtins import str
from builtins import object
from past.utils import old_div

import sys
import time
import itertools
from functools import wraps

import py.test

import mdp
from mdp import numx, numx_rand, numx_fft, numx_linalg, utils
from numpy.testing import (assert_array_equal, assert_array_almost_equal,
                           assert_equal, assert_almost_equal)

mean = numx.mean
std = numx.std
normal = mdp.numx_rand.normal
uniform = mdp.numx_rand.random
testtypes = [numx.dtype('d'), numx.dtype('f')]
testtypeschar = [t.char for t in testtypes]
testdecimals = {testtypes[0]: 12, testtypes[1]: 6}
decimal = 7
mult = mdp.utils.mult

#### test tools
def assert_array_almost_equal_diff(x,y,digits,err_msg=''):
    x,y = numx.asarray(x), numx.asarray(y)
    msg = '\nArrays are not almost equal'
    assert 0 in [len(numx.shape(x)),len(numx.shape(y))] \
           or (len(numx.shape(x))==len(numx.shape(y)) and \
               numx.alltrue(numx.equal(numx.shape(x),numx.shape(y)))),\
               msg + ' (shapes %s, %s mismatch):\n\t' \
               % (numx.shape(x),numx.shape(y)) + err_msg
    maxdiff = old_div(max(numx.ravel(abs(x-y))),\
              max(max(abs(numx.ravel(x))),max(abs(numx.ravel(y)))))
    if numx.iscomplexobj(x) or numx.iscomplexobj(y): maxdiff = old_div(maxdiff,2)
    cond =  maxdiff< 10**(-digits)
    msg = msg+'\n\t Relative maximum difference: %e'%(maxdiff)+'\n\t'+\
          'Array1: '+str(x)+'\n\t'+\
          'Array2: '+str(y)+'\n\t'+\
          'Absolute Difference: '+str(abs(y-x))
    assert cond, msg

def assert_type_equal(act, des):
    assert act == numx.dtype(des), \
           'dtype mismatch: "%s" (should be "%s") '%(act,des)

def get_random_mix(mat_dim = None, type = "d", scale = 1,\
                    rand_func = uniform, avg = 0, \
                    std_dev = 1):
    if mat_dim is None:
        mat_dim = (500, 5)
    T = mat_dim[0]
    N = mat_dim[1]
    d = 0
    while d < 1E-3:
        #mat = ((rand_func(size=mat_dim)-0.5)*scale).astype(type)
        mat = rand_func(size=(T,N)).astype(type)
        # normalize
        mat -= mean(mat,axis=0)
        mat /= std(mat,axis=0)
        # check that the minimum eigenvalue is finite and positive
        d1 = min(mdp.utils.symeig(mdp.utils.mult(mat.T, mat), eigenvectors = 0))
        if std_dev is not None: mat *= std_dev
        if avg is not None: mat += avg
        mix = (rand_func(size=(N,N))*scale).astype(type)
        matmix = mdp.utils.mult(mat,mix)
        matmix_n = matmix - mean(matmix, axis=0)
        matmix_n /= std(matmix_n, axis=0)
        d2 = min(mdp.utils.symeig(mdp.utils.mult(matmix_n.T,matmix_n),
                                  eigenvectors=0))
        d = min(d1, d2)
    return mat, mix, matmix

def verify_ICANode(icanode, rand_func = uniform, vars=3, N=8000, prec=3):
    dim = (N, vars)
    mat,mix,inp = get_random_mix(rand_func=rand_func,mat_dim=dim)
    icanode.train(inp)
    act_mat = icanode.execute(inp)
    cov = mdp.utils.cov2(old_div((mat-mean(mat,axis=0)),std(mat,axis=0)), act_mat)
    maxima = numx.amax(abs(cov), axis=0)
    assert_array_almost_equal(maxima,numx.ones(vars), prec)

def verify_ICANodeMatrices(icanode, rand_func=uniform, vars=3, N=8000):
    dim = (N, vars)
    mat,mix,inp = get_random_mix(rand_func=rand_func, mat_dim=dim, avg=0)
    icanode.train(inp)
    # test projection matrix
    act_mat = icanode.execute(inp)
    T = icanode.get_projmatrix()
    exp_mat = mdp.utils.mult(inp, T)
    assert_array_almost_equal(act_mat,exp_mat,6)
    # test reconstruction matrix
    out = act_mat.copy()
    act_mat = icanode.inverse(out)
    B = icanode.get_recmatrix()
    exp_mat = mdp.utils.mult(out, B)
    assert_array_almost_equal(act_mat,exp_mat,6)

class BogusNode(mdp.Node):
    @staticmethod
    def is_trainable(): return False
    def _execute(self,x): return 2*x
    def _inverse(self,x): return 0.5*x

class BogusNodeTrainable(mdp.Node):
    def _train(self, x):
        pass
    def _stop_training(self):
        self.bogus_attr = 1

class BogusExceptNode(mdp.Node):
    def _train(self,x):
        self.bogus_attr = 1
        raise Exception("Bogus Exception")

    def _execute(self,x):
        raise Exception("Bogus Exception")

class BogusMultiNode(mdp.Node):

    def __init__(self):
        super(BogusMultiNode, self).__init__()
        self.visited = []

    def _get_train_seq(self):
        return [(self.train1, self.stop1),
                (self.train2, self.stop2)]

    def train1(self, x):
        self.visited.append(1)
    def stop1(self):
        self.visited.append(2)
    def train2(self, x):
        self.visited.append(3)
    def stop2(self):
        self.visited.append(4)

#_spinner = itertools.cycle((' /\b\b', ' -\b\b', ' \\\b\b', ' |\b\b'))
_spinner = itertools.cycle((' .\b\b', ' o\b\b', ' 0\b\b', ' O\b\b',
                            ' 0\b\b', ' o\b\b'))
#_spinner = itertools.cycle([" '\b\b"]*2 + [' !\b\b']*2 + [' .\b\b']*2 +
#                           [' !\b\b']*2)

def spinner():
    sys.stderr.write(next(_spinner))
    sys.stderr.flush()

class skip_on_condition(object):
    """Skip a test if the eval(condition_str, namespace) returns True.

    namespace contains sys, os, and the mdp module.
    """
    def __init__(self, condition_str, skipping_msg=None):
        self.condition_str = condition_str
        if skipping_msg is None:
            self.skipping_msg = "Condition %s not met." % condition_str
        else:
            self.skipping_msg = skipping_msg

    def __call__(self, f):
        import sys, os
        @wraps(f)
        def wrapped_f(*args, **kwargs):
            namespace = {'sys': sys,
                         'os': os,
                         'mdp': mdp}
            if eval(self.condition_str, namespace):
                py.test.skip(self.skipping_msg)
            f(*args, **kwargs)
        return wrapped_f




from bimdp import BiNode, MSG_ID_SEP
from bimdp.nodes import IdentityBiNode


class SenderBiNode(IdentityBiNode):
    """Sends the incoming x data to another node via bi_message."""

    def __init__(self, recipient_id=None, **kwargs):
        """Initialize the internal variables.

        recipient_id -- None or the id for the data recipient.
        """
        super(SenderBiNode, self).__init__(**kwargs)
        self._recipient_id = recipient_id

    def _execute(self, x, no_x=None):
        """Add msg_x to the message (adressed to a target if defined)."""
        msg = dict()
        if self._recipient_id:
            msg[self._recipient_id + MSG_ID_SEP + "msg_x"] = x
        else:
            msg["msg_x"] = x
        if no_x:
            x = None
        return x, msg



from __future__ import division
from future import standard_library
standard_library.install_aliases()
from builtins import str
from builtins import range
from past.utils import old_div
from builtins import object

__docformat__ = "restructuredtext en"

import mdp
from mdp import numx, utils, Node, NodeException, PreserveDimNode

import pickle as pickle
import pickle as real_pickle

class IdentityNode(PreserveDimNode):
    """Execute returns the input data and the node is not trainable.

    This node can be instantiated and is for example useful in
    complex network layouts.
    """

    def _get_supported_dtypes(self):
        """Return the list of dtypes supported by this node."""
        return (mdp.utils.get_dtypes('AllFloat') +
                mdp.utils.get_dtypes('AllInteger') +
                mdp.utils.get_dtypes('Character'))

    @staticmethod
    def is_trainable():
        return False

class OneDimensionalHitParade(object):
    """
    Class to produce hit-parades (i.e., a list of the largest
    and smallest values) out of a one-dimensional time-series.
    """
    
    def __init__(self, n, d, real_dtype="d", integer_dtype="l"):
        """
        Input arguments:
        n -- Number of maxima and minima to remember
        d -- Minimum gap between two hits

        real_dtype -- dtype of sequence items
        integer_dtype -- dtype of sequence indices
        Note: be careful with dtypes!
        """
        self.n = int(n)
        self.d = int(d)
        self.iM = numx.zeros((n, ), dtype=integer_dtype)
        self.im = numx.zeros((n, ), dtype=integer_dtype)
        
        real_dtype = numx.dtype(real_dtype)
        if real_dtype in mdp.utils.get_dtypes('AllInteger'):
            max_num = numx.iinfo(real_dtype).max
            min_num = numx.iinfo(real_dtype).min
        else:
            max_num = numx.finfo(real_dtype).max
            min_num = numx.finfo(real_dtype).min
        self.M = numx.array([min_num]*n, dtype=real_dtype)
        self.m = numx.array([max_num]*n, dtype=real_dtype)
        
        self.lM = 0
        self.lm = 0

    def update(self, inp):
        """
        Input arguments:
        inp -- tuple (time-series, time-indices)
        """
        (x, ix) = inp
        rows = len(x)
        d = self.d
        M = self.M
        m = self.m
        iM = self.iM
        im = self.im
        lM = self.lM
        lm = self.lm
        for i in range(rows):
            k1 = M.argmin()
            k2 = m.argmax()
            if x[i] > M[k1]:
                if ix[i]-iM[lM] <= d and x[i] > M[lM]:
                    M[lM] = x[i]
                    iM[lM] = ix[i]
                elif ix[i]-iM[lM] > d:
                    M[k1] = x[i]
                    iM[k1] = ix[i]
                    lM = k1
            if x[i] < m[k2]:
                if ix[i]-im[lm] <= d and x[i] < m[lm]:
                    m[lm] = x[i]
                    im[lm] = ix[i]
                elif ix[i]-im[lm] > d:
                    m[k2] = x[i]
                    im[k2] = ix[i]
                    lm = k2
        self.M = M
        self.m = m
        self.iM = iM
        self.im = im
        self.lM = lM
        self.lm = lm

    def get_maxima(self):
        """
        Return the tuple (maxima, time-indices).
        Maxima are sorted in descending order.
        """
        iM = self.iM
        M = self.M
        sort = M.argsort()
        return M[sort[::-1]], iM[sort[::-1]]

    def get_minima(self):
        """
        Return the tuple (minima, time-indices).
        Minima are sorted in ascending order.
        """
        im = self.im
        m = self.m
        sort = m.argsort()
        return m[sort], im[sort]


class HitParadeNode(PreserveDimNode):
    """Collect the first ``n`` local maxima and minima of the training signal
    which are separated by a minimum gap ``d``.

    This is an analysis node, i.e. the data is analyzed during training
    and the results are stored internally. Use the
    ``get_maxima`` and ``get_minima`` methods to access them.
    """

    def __init__(self, n, d=1, input_dim=None, output_dim=None, dtype=None):
        """
        Input arguments:
        n -- Number of maxima and minima to store
        d -- Minimum gap between two maxima or two minima
        """
        super(HitParadeNode, self).__init__(input_dim=input_dim,
                                            output_dim=output_dim,
                                            dtype=dtype)
        self.n = int(n)
        self.d = int(d)
        self.itype = 'int64'
        self.hit = None
        self.tlen = 0

    def _set_input_dim(self, n):
        self._input_dim = n
        self.output_dim = n

    def _get_supported_dtypes(self):
        """Return the list of dtypes supported by this node."""
        return (mdp.utils.get_dtypes('Float') +
                mdp.utils.get_dtypes('AllInteger'))

    def _train(self, x):
        hit = self.hit
        old_tlen = self.tlen
        if hit is None:
            hit = [OneDimensionalHitParade(self.n, self.d, self.dtype,
                                           self.itype)
                   for c in range(self.input_dim)]
        tlen = old_tlen + x.shape[0]
        indices = numx.arange(old_tlen, tlen)
        for c in range(self.input_dim):
            hit[c].update((x[:, c], indices))
        self.hit = hit
        self.tlen = tlen

    def get_maxima(self):
        """
        Return the tuple (maxima, indices).
        Maxima are sorted in descending order.

        If the training phase has not been completed yet, call
        stop_training.
        """
        self._if_training_stop_training()
        cols = self.input_dim
        n = self.n
        hit = self.hit
        iM = numx.zeros((n, cols), dtype=self.itype)
        M = numx.ones((n, cols), dtype=self.dtype)
        for c in range(cols):
            M[:, c], iM[:, c] = hit[c].get_maxima()
        return M, iM

    def get_minima(self):
        """
        Return the tuple (minima, indices).
        Minima are sorted in ascending order.

        If the training phase has not been completed yet, call
        stop_training.
        """
        self._if_training_stop_training()
        cols = self.input_dim
        n = self.n
        hit = self.hit
        im = numx.zeros((n, cols), dtype=self.itype)
        m = numx.ones((n, cols), dtype=self.dtype)
        for c in range(cols):
            m[:, c], im[:, c] = hit[c].get_minima()
        return m, im

class TimeFramesNode(Node):
    """Copy delayed version of the input signal on the space dimensions.

    For example, for ``time_frames=3`` and ``gap=2``::

      [ X(1) Y(1)        [ X(1) Y(1) X(3) Y(3) X(5) Y(5)
        X(2) Y(2)          X(2) Y(2) X(4) Y(4) X(6) Y(6)
        X(3) Y(3)   -->    X(3) Y(3) X(5) Y(5) X(7) Y(7)
        X(4) Y(4)          X(4) Y(4) X(6) Y(6) X(8) Y(8)
        X(5) Y(5)          ...  ...  ...  ...  ...  ... ]
        X(6) Y(6)
        X(7) Y(7)
        X(8) Y(8)
        ...  ...  ]

    It is not always possible to invert this transformation (the
    transformation is not surjective. However, the ``pseudo_inverse``
    method does the correct thing when it is indeed possible.
    """

    def __init__(self, time_frames, gap=1,
                 input_dim=None, dtype=None):
        """
        Input arguments:
        time_frames -- Number of delayed copies
        gap -- Time delay between the copies
        """
        self.time_frames = time_frames
        super(TimeFramesNode, self).__init__(input_dim=input_dim,
                                             output_dim=None,
                                             dtype=dtype)
        self.gap = gap

    def _get_supported_dtypes(self):
        """Return the list of dtypes supported by this node."""
        return (mdp.utils.get_dtypes('AllFloat') +
                mdp.utils.get_dtypes('AllInteger') +
                mdp.utils.get_dtypes('Character'))

    @staticmethod
    def is_trainable():
        return False

    @staticmethod
    def is_invertible():
        return False

    def _set_input_dim(self, n):
        self._input_dim = n
        self._output_dim = n*self.time_frames

    def _set_output_dim(self, n):
        msg = 'Output dim can not be explicitly set!'
        raise NodeException(msg)

    def _execute(self, x):
        gap = self.gap
        tf = x.shape[0] - (self.time_frames-1)*gap
        rows = self.input_dim
        cols = self.output_dim
        y = numx.zeros((tf, cols), dtype=self.dtype)
        for frame in range(self.time_frames):
            y[:, frame*rows:(frame+1)*rows] = x[gap*frame:gap*frame+tf, :]
        return y

    def pseudo_inverse(self, y):
        """This function returns a pseudo-inverse of the execute frame.
        y == execute(x) only if y belongs to the domain of execute and
        has been computed with a sufficently large x.
        If gap > 1 some of the last rows will be filled with zeros.
        """

        self._if_training_stop_training()

        # set the output dimension if necessary
        if not self.output_dim:
            # if the input_dim is not defined, raise an exception
            if not self.input_dim:
                errstr = ("Number of input dimensions undefined. Inversion"
                          "not possible.")
                raise NodeException(errstr)
            self.outputdim = self.input_dim

        # control the dimension of y
        self._check_output(y)
        # cast
        y = self._refcast(y)

        gap = self.gap
        exp_length = y.shape[0]
        cols = self.input_dim
        rest = (self.time_frames-1)*gap
        rows = exp_length + rest
        x = numx.zeros((rows, cols), dtype=self.dtype)
        x[:exp_length, :] = y[:, :cols]
        count = 1
        # Note that if gap > 1 some of the last rows will be filled with zeros!
        block_sz = min(gap, exp_length)
        for row in range(max(exp_length, gap), rows, gap):
            x[row:row+block_sz, :] = y[-block_sz:, count*cols:(count+1)*cols]
            count += 1
        return x

class TimeDelayNode(TimeFramesNode):
    """
    Copy delayed version of the input signal on the space dimensions.

    For example, for ``time_frames=3`` and ``gap=2``::

      [ X(1) Y(1)        [ X(1) Y(1)   0    0    0    0
        X(2) Y(2)          X(2) Y(2)   0    0    0    0
        X(3) Y(3)   -->    X(3) Y(3) X(1) Y(1)   0    0
        X(4) Y(4)          X(4) Y(4) X(2) Y(2)   0    0
        X(5) Y(5)          X(5) Y(5) X(3) Y(3) X(1) Y(1)
        X(6) Y(6)          ...  ...  ...  ...  ...  ... ]
        X(7) Y(7)
        X(8) Y(8)
        ...  ...  ]

    This node provides similar functionality as the ``TimeFramesNode``, only
    that it performs a time embedding into the past rather than into the future.

    See ``TimeDelaySlidingWindowNode`` for a sliding window delay node for
    application in a non-batch manner.

    Original code contributed by Sebastian Hoefer.
    Dec 31, 2010
    """

    def __init__(self, time_frames, gap=1, input_dim=None, dtype=None):
        """
        Input arguments:
        time_frames -- Number of delayed copies
        gap -- Time delay between the copies
        """
        super(TimeDelayNode, self).__init__(time_frames, gap,
                                            input_dim, dtype)

    def _execute(self, x):
        gap = self.gap
        rows = x.shape[0]
        cols = self.output_dim
        n = self.input_dim

        y = numx.zeros((rows, cols), dtype=self.dtype)

        for frame in range(self.time_frames):
            y[gap*frame:, frame*n:(frame+1)*n] = x[:rows-gap*frame, :]

        return y

    def pseudo_inverse(self, y):
        raise NotImplementedError

class TimeDelaySlidingWindowNode(TimeDelayNode):
    """
    ``TimeDelaySlidingWindowNode`` is an alternative to ``TimeDelayNode``
    which should be used for online learning/execution. Whereas the
    ``TimeDelayNode`` works in a batch manner, for online application
    a sliding window is necessary which yields only one row per call.

    Applied to the same data the collection of all returned rows of the
    ``TimeDelaySlidingWindowNode`` is equivalent to the result of the
    ``TimeDelayNode``.

    Original code contributed by Sebastian Hoefer.
    Dec 31, 2010
    """
    def __init__(self, time_frames, gap=1, input_dim=None, dtype=None):
        """
        Input arguments:
        time_frames -- Number of delayed copies
        gap -- Time delay between the copies
        """

        self.time_frames = time_frames
        self.gap = gap
        super(TimeDelaySlidingWindowNode, self).__init__(time_frames, gap,
                                                         input_dim, dtype)
        self.sliding_wnd = None
        self.cur_idx = 0
        self.slide = False

    def _init_sliding_window(self):
        rows = self.gap+1
        cols = self.input_dim*self.time_frames
        self.sliding_wnd = numx.zeros((rows, cols), dtype=self.dtype)

    def _execute(self, x):
        assert x.shape[0] == 1

        if self.sliding_wnd == None:
            self._init_sliding_window()

        gap = self.gap
        rows = self.sliding_wnd.shape[0]
        cols = self.output_dim
        n = self.input_dim

        new_row = numx.zeros(cols, dtype=self.dtype)
        new_row[:n] = x

        # Slide
        if self.slide:
            self.sliding_wnd[:-1, :] = self.sliding_wnd[1:, :]

        # Delay
        if self.cur_idx-gap >= 0:
            new_row[n:] = self.sliding_wnd[self.cur_idx-gap, :-n]

        # Add new row to matrix
        self.sliding_wnd[self.cur_idx, :] = new_row

        if self.cur_idx < rows-1:
            self.cur_idx = self.cur_idx+1 
        else:
            self.slide = True

        return new_row[numx.newaxis,:]

class EtaComputerNode(Node):
    """Compute the eta values of the normalized training data.

    The delta value of a signal is a measure of its temporal
    variation, and is defined as the mean of the derivative squared,
    i.e. ``delta(x) = mean(dx/dt(t)^2)``.  ``delta(x)`` is zero if
    ``x`` is a constant signal, and increases if the temporal variation
    of the signal is bigger.

    The eta value is a more intuitive measure of temporal variation,
    defined as::
    
       eta(x) = T/(2*pi) * sqrt(delta(x))

    If ``x`` is a signal of length ``T`` which consists of a sine function
    that accomplishes exactly ``N`` oscillations, then ``eta(x)=N``.

    ``EtaComputerNode`` normalizes the training data to have unit
    variance, such that it is possible to compare the temporal
    variation of two signals independently from their scaling.

    Reference: Wiskott, L. and Sejnowski, T.J. (2002).
    Slow Feature Analysis: Unsupervised Learning of Invariances,
    Neural Computation, 14(4):715-770.

    Important: if a data chunk is tlen data points long, this node is
    going to consider only the first tlen-1 points together with their
    derivatives. This means in particular that the variance of the
    signal is not computed on all data points. This behavior is
    compatible with that of ``SFANode``.

    This is an analysis node, i.e. the data is analyzed during training
    and the results are stored internally.  Use the method
    ``get_eta`` to access them.
    """

    def __init__(self, input_dim=None, dtype=None):
        super(EtaComputerNode, self).__init__(input_dim, None, dtype)
        self._initialized = 0

    def _set_input_dim(self, n):
        self._input_dim = n
        self.output_dim = n

    def _init_internals(self):
        input_dim = self.input_dim
        self._mean = numx.zeros((input_dim,), dtype='d')
        self._var = numx.zeros((input_dim,), dtype='d')
        self._tlen = 0
        self._diff2 = numx.zeros((input_dim,), dtype='d')
        self._initialized = 1

    def _train(self, data):
        # here SignalNode.train makes an automatic refcast
        if not self._initialized:
            self._init_internals()

        rdata = data[:-1]
        self._mean += rdata.sum(axis=0)
        self._var += (rdata*rdata).sum(axis=0)
        self._tlen += rdata.shape[0]
        td_data = utils.timediff(data)
        self._diff2 += (td_data*td_data).sum(axis=0)

    def _stop_training(self):
        var_tlen = self._tlen-1
        # unbiased
        var = old_div((self._var - self._mean*self._mean/self._tlen),var_tlen)

        # biased
        #var = (self._var - self._mean*self._mean/self._tlen)/self._tlen

        # old formula: wrong! is neither biased nor unbiased
        #var = (self._var/var_tlen) - (self._mean/self._tlen)**2

        self._var = var
        delta = old_div((old_div(self._diff2,self._tlen)),var)
        self._delta = delta
        self._eta = old_div(numx.sqrt(delta),(2*numx.pi))

    def get_eta(self, t=1):
        """Return the eta values of the data received during the training
        phase. If the training phase has not been completed yet, call
        stop_training.

        :Arguments:
           t
             Sampling frequency in Hz.

             The original definition in (Wiskott and Sejnowski, 2002)
             is obtained for ``t=self._tlen``, while for ``t=1`` (default),
             this corresponds to the beta-value defined in
             (Berkes and Wiskott, 2005).
        """
        self._if_training_stop_training()
        return self._refcast(self._eta*t)


class NoiseNode(PreserveDimNode):
    """Inject multiplicative or additive noise into the input data.

    Original code contributed by Mathias Franzius.
    """

    def __init__(self, noise_func=mdp.numx_rand.normal, noise_args=(0, 1),
                 noise_type='additive',
                 input_dim=None, output_dim=None, dtype=None):
        """
        Add noise to input signals.

        :Arguments:
          noise_func
            A function that generates noise. It must
            take a ``size`` keyword argument and return
            a random array of that size. Default is normal noise.

          noise_args
            Tuple of additional arguments passed to `noise_func`.
            Default is (0,1) for (mean, standard deviation)
            of the normal distribution.

          noise_type
            Either ``'additive'`` or ``'multiplicative'``.

            'additive'
               returns ``x + noise``.
            'multiplicative'
               returns ``x * (1 + noise)``

            Default is ``'additive'``.
        """
        super(NoiseNode, self).__init__(input_dim=input_dim,
                                        output_dim=output_dim,
                                        dtype=dtype)
        self.noise_func = noise_func
        self.noise_args = noise_args
        valid_noise_types = ['additive', 'multiplicative']
        if noise_type not in valid_noise_types:
            err_str = '%s is not a valid noise type' % str(noise_type)
            raise NodeException(err_str)
        else:
            self.noise_type = noise_type

    def _get_supported_dtypes(self):
        """Return the list of dtypes supported by this node."""
        return (mdp.utils.get_dtypes('Float') +
                mdp.utils.get_dtypes('AllInteger'))

    @staticmethod
    def is_trainable():
        return False

    @staticmethod
    def is_invertible():
        return False

    def _execute(self, x):
        noise_mat = self._refcast(self.noise_func(*self.noise_args,
                                                  **{'size': x.shape}))
        if self.noise_type == 'additive':
            return x+noise_mat
        elif self.noise_type == 'multiplicative':
            return x*(1.+noise_mat)

    def save(self, filename, protocol = -1):
        """Save a pickled serialization of the node to 'filename'.
        If 'filename' is None, return a string.

        Note: the pickled Node is not guaranteed to be upward or
        backward compatible."""
        if filename is None:
            # cPickle seems to create an error, probably due to the
            # self.noise_func attribute.
            return real_pickle.dumps(self, protocol)
        else:
            # if protocol != 0 open the file in binary mode
            mode = 'w' if protocol == 0 else 'wb'
            with open(filename, mode) as flh:
                real_pickle.dump(self, flh, protocol)


class NormalNoiseNode(PreserveDimNode):
    """Special version of ``NoiseNode`` for Gaussian additive noise.

    Unlike ``NoiseNode`` it does not store a noise function reference but simply
    uses ``numx_rand.normal``.
    """

    def __init__(self, noise_args=(0, 1),
                 input_dim=None, output_dim=None, dtype=None):
        """Set the noise parameters.

        noise_args -- Tuple of (mean, standard deviation) for the normal
            distribution, default is (0,1).
        """
        super(NormalNoiseNode, self).__init__(input_dim=input_dim,
                                              output_dim=output_dim,
                                              dtype=dtype)
        self.noise_args = noise_args

    @staticmethod
    def is_trainable():
        return False

    @staticmethod
    def is_invertible():
        return False

    def _execute(self, x):
        noise = self._refcast(mdp.numx_rand.normal(size=x.shape) *
                                    self.noise_args[1]
                              + self.noise_args[0])
        return x + noise


class CutoffNode(PreserveDimNode):
    """Node to cut off values at specified bounds.

    Works similar to ``numpy.clip``, but also works when only a lower or upper
    bound is specified.
    """

    def __init__(self, lower_bound=None, upper_bound=None,
                 input_dim=None, output_dim=None, dtype=None):
        """Initialize node.

        :Parameters:
          lower_bound
            Data values below this are cut to the ``lower_bound`` value.
            If ``lower_bound`` is ``None`` no cutoff is performed.
          upper_bound
            Works like ``lower_bound``.
        """
        super(CutoffNode, self).__init__(input_dim=input_dim,
                                         output_dim=output_dim,
                                         dtype=dtype)
        self.lower_bound = lower_bound
        self.upper_bound = upper_bound

    @staticmethod
    def is_trainable():
        return False

    @staticmethod
    def is_invertible():
        return False

    def _get_supported_dtypes(self):
        return (mdp.utils.get_dtypes('Float') +
                mdp.utils.get_dtypes('AllInteger'))

    def _execute(self, x):
        """Return the clipped data."""
        # n.clip() does not work, since it does not accept None for one bound
        if self.lower_bound is not None:
            x = numx.where(x >= self.lower_bound, x, self.lower_bound)
        if self.upper_bound is not None:
            x = numx.where(x <= self.upper_bound, x, self.upper_bound)
        return x


class HistogramNode(PreserveDimNode):
    """Node which stores a history of the data during its training phase.

    The data history is stored in ``self.data_hist`` and can also be deleted to
    free memory. Alternatively it can be automatically pickled to disk.

    Note that data is only stored during training.
    """

    def __init__(self, hist_fraction=1.0, hist_filename=None,
                 input_dim=None, output_dim=None, dtype=None):
        """Initialize the node.

        hist_fraction -- Defines the fraction of the data that is stored
            randomly.
        hist_filename -- Filename for the file to which the data history will
            be pickled after training. The data is pickled when stop_training
            is called and data_hist is then cleared (to free memory).
            If filename is None (default value) then data_hist is not cleared
            and can be directly used after training.
        """
        super(HistogramNode, self).__init__(input_dim=input_dim,
                                            output_dim=output_dim,
                                            dtype=dtype)
        self._hist_filename = hist_filename
        self.hist_fraction = hist_fraction
        self.data_hist = None  # stores the data history

    def _get_supported_dtypes(self):
        return (mdp.utils.get_dtypes('AllFloat') +
                mdp.utils.get_dtypes('AllInteger') +
                mdp.utils.get_dtypes('Character'))

    def _train(self, x):
        """Store the history data."""
        if self.hist_fraction < 1.0:
            x = x[numx.random.random(len(x)) < self.hist_fraction]
        if self.data_hist is not None:
            self.data_hist = numx.concatenate([self.data_hist, x])
        else:
            self.data_hist = x

    def _stop_training(self):
        """Pickle the histogram data to file and clear it if required."""
        super(HistogramNode, self)._stop_training()
        if self._hist_filename:
            pickle_file = open(self._hist_filename, "wb")
            try:
                pickle.dump(self.data_hist, pickle_file, protocol=-1)
            finally:
                pickle_file.close( )
            self.data_hist = None

class AdaptiveCutoffNode(HistogramNode):
    """Node which uses the data history during training to learn cutoff values.

    As opposed to the simple ``CutoffNode``, a different cutoff value is learned
    for each data coordinate. For example if an upper cutoff fraction of
    0.05 is specified, then the upper cutoff bound is set so that the upper
    5% of the training data would have been clipped (in each dimension).
    The cutoff bounds are then applied during execution.
    This node also works as a ``HistogramNode``, so the histogram data is stored.

    When ``stop_training`` is called the cutoff values for each coordinate are
    calculated based on the collected histogram data.
    """

    def __init__(self, lower_cutoff_fraction=None, upper_cutoff_fraction=None,
                 hist_fraction=1.0, hist_filename=None,
                 input_dim=None, output_dim=None, dtype=None):
        """Initialize the node.

        :Parameters:
          lower_cutoff_fraction
            Fraction of data that will be cut off after
            the training phase (assuming the data distribution does not
            change). If set to ``None`` (default value) no cutoff is performed.
          upper_cutoff_fraction
            Works like `lower_cutoff_fraction`.
          hist_fraction
            Defines the fraction of the data that is stored for the
            histogram.
          hist_filename
            Filename for the file to which the data history will be
            pickled after training. The data is pickled when
            `stop_training` is called and ``data_hist`` is then
            cleared (to free memory).  If filename is ``None``
            (default value) then ``data_hist`` is not cleared and can
            be directly used after training.
        """
        super(AdaptiveCutoffNode, self).__init__(hist_fraction=hist_fraction,
                                                 hist_filename=hist_filename,
                                                 input_dim=input_dim,
                                                 output_dim=output_dim,
                                                 dtype=dtype)
        self.lower_cutoff_fraction = lower_cutoff_fraction
        self.upper_cutoff_fraction = upper_cutoff_fraction
        self.lower_bounds = None
        self.upper_bounds = None
        
    def _get_supported_dtypes(self):
        return (mdp.utils.get_dtypes('Float') +
                mdp.utils.get_dtypes('AllInteger'))

    def _stop_training(self):
        """Calculate the cutoff bounds based on collected histogram data."""
        if self.lower_cutoff_fraction or self.upper_cutoff_fraction:
            sorted_data = self.data_hist.copy()
            sorted_data.sort(axis=0)
            if self.lower_cutoff_fraction:
                index = self.lower_cutoff_fraction * len(sorted_data)
                self.lower_bounds = sorted_data[int(index)]
            if self.upper_cutoff_fraction:
                index = (len(sorted_data) -
                         self.upper_cutoff_fraction * len(sorted_data))
                self.upper_bounds = sorted_data[int(index)]
        super(AdaptiveCutoffNode, self)._stop_training()

    def _execute(self, x):
        """Return the clipped data."""
        if self.lower_bounds is not None:
            x = numx.where(x >= self.lower_bounds, x, self.lower_bounds)
        if self.upper_bounds is not None:
            x = numx.where(x <= self.upper_bounds, x, self.upper_bounds)
        return x



from __future__ import print_function
from __future__ import division
from builtins import range
from past.utils import old_div
__docformat__ = "restructuredtext en"

import mdp
from .ica_nodes import ICANode
numx, numx_rand, numx_linalg = mdp.numx, mdp.numx_rand, mdp.numx_linalg

mult = mdp.utils.mult

class JADENode(ICANode):
    """
    Perform Independent Component Analysis using the JADE algorithm.
    Note that JADE is a batch-algorithm. This means that it needs
    all input data before it can start and compute the ICs.
    The algorithm is here given as a Node for convenience, but it
    actually accumulates all inputs it receives. Remember that to avoid
    running out of memory when you have many components and many time samples.

    JADE does not support the telescope mode.

    Main references:
    
      * Cardoso, Jean-Francois and Souloumiac, Antoine (1993).
        Blind beamforming for non Gaussian signals.
        Radar and Signal Processing, IEE Proceedings F, 140(6): 362-370.
      * Cardoso, Jean-Francois (1999).
        High-order contrasts for independent component analysis.
        Neural Computation, 11(1): 157-192.

    Original code contributed by: 
    Gabriel Beckers (2008).

    History:
    
    - May 2005    version 1.8 for MATLAB released by Jean-Francois Cardoso
    - Dec 2007    MATLAB version 1.8 ported to Python/NumPy by Gabriel Beckers
    - Feb 15 2008 Python/NumPy version adapted for MDP by Gabriel Beckers
    """

    def __init__(self, limit = 0.001, max_it=1000, verbose = False,
                 whitened = False, white_comp = None, white_parm = None,
                 input_dim = None, dtype = None):
        """
        Input arguments:

        General:

        whitened -- Set whitened == True if input data are already whitened.
                    Otherwise the node will whiten the data itself

        white_comp -- If whitened == False, you can set 'white_comp' to the
                      number of whitened components to keep during the
                      calculation (i.e., the input dimensions are reduced to
                      white_comp by keeping the components of largest variance).

        white_parm -- a dictionary with additional parameters for whitening.
                      It is passed directly to the WhiteningNode constructor.
                      Ex: white_parm = { 'svd' : True }

        limit -- convergence threshold.

        Specific for JADE:

        max_it -- maximum number of iterations

        """
        super(JADENode, self).__init__(limit, False, verbose, whitened,
                                       white_comp, white_parm, input_dim,
                                       dtype)

        self.max_it = max_it

    def core(self, data):
        # much of the code here is a more or less line by line translation of
        # the original matlab code by Jean-Francois Cardoso.
        append = numx.append
        arange = numx.arange
        arctan2 = numx.arctan2
        array = numx.array
        concatenate = numx.concatenate
        cos = numx.cos
        sin = numx.sin
        sqrt = numx.sqrt
        dtype = self.dtype
        verbose = self.verbose
        max_it = self.max_it
        (T, m) = data.shape
        X = data

        if verbose:
            print("jade -> Estimating cumulant matrices")

        # Dim. of the space of real symm matrices
        dimsymm = (m*(m+1)) // 2
        # number of cumulant matrices
        nbcm = dimsymm
        # Storage for cumulant matrices
        CM = numx.zeros((m, m*nbcm), dtype=dtype)
        R = numx.eye(m, dtype=dtype)
        # Temp for a cum. matrix
        Qij = numx.zeros((m, m), dtype=dtype)
        # Temp
        Xim = numx.zeros(m, dtype=dtype)
        # Temp
        Xijm = numx.zeros(m, dtype=dtype)

        # I am using a symmetry trick to save storage. I should write a short
        # note one of these days explaining what is going on here.
        # will index the columns of CM where to store the cum. mats.
        Range = arange(m)

        for im in range(m):
            Xim = X[:, im]
            Xijm = Xim*Xim
            # Note to myself: the -R on next line can be removed: it does not
            # affect the joint diagonalization criterion
            Qij = ( old_div(mult(Xijm*X.T, X), float(T))
                    - R - 2 * numx.outer(R[:,im], R[:,im]) )
            CM[:, Range] = Qij
            Range += m
            for jm in range(im):
                Xijm = Xim*X[:, jm]
                Qij = ( sqrt(2) * mult(Xijm*X.T, X) / T
                        - numx.outer(R[:,im], R[:,jm]) - numx.outer(R[:,jm],
                                                                    R[:,im]) )
                CM[:, Range] = Qij
                Range += m

        # Now we have nbcm = m(m+1)/2 cumulants matrices stored in a big
        # m x m*nbcm array.

        # Joint diagonalization of the cumulant matrices
        # ==============================================

        V = numx.eye(m, dtype=dtype)

        Diag = numx.zeros(m, dtype=dtype)
        On = 0.0
        Range = arange(m)
        for im in range(nbcm):
            Diag = numx.diag(CM[:, Range])
            On = On + (Diag*Diag).sum(axis=0)
            Range += m

        Off = (CM*CM).sum(axis=0) - On
        # A statistically scaled threshold on `small" angles
        seuil = old_div((self.limit*self.limit), sqrt(T))
        # sweep number
        encore = True
        sweep = 0
        # Total number of rotations
        updates = 0
        # Number of rotations in a given seep
        upds = 0
        g = numx.zeros((2, nbcm), dtype=dtype)
        gg = numx.zeros((2, 2), dtype=dtype)
        G = numx.zeros((2, 2), dtype=dtype)
        c = 0
        s = 0
        ton = 0
        toff = 0
        theta = 0
        Gain = 0


        # Joint diagonalization proper
        # ============================
        if verbose:
            print("jade -> Contrast optimization by joint diagonalization")

        while encore:
            encore = False
            if verbose:
                print("jade -> Sweep #%3d" % sweep, end=' ')
            sweep += 1
            upds  = 0

            for p in range(m-1):
                for q in range(p+1, m):

                    Ip = arange(p, m*nbcm, m)
                    Iq = arange(q, m*nbcm, m)

                    # computation of Givens angle
                    g = concatenate([numx.atleast_2d(CM[p, Ip] - CM[q, Iq]),
                                     numx.atleast_2d(CM[p, Iq] + CM[q, Ip])])
                    gg = mult(g, g.T)
                    ton = gg[0, 0] - gg[1, 1]
                    toff = gg[0, 1] + gg[1, 0]
                    theta = 0.5 * arctan2(toff, ton + sqrt(ton*ton+toff*toff))
                    Gain = old_div((sqrt(ton * ton + toff * toff) - ton), 4.0)

                    # Givens update
                    if abs(theta) > seuil:
                        encore = True
                        upds = upds + 1
                        c = cos(theta)
                        s = sin(theta)
                        G = array([[c, -s] , [s, c] ])
                        pair = array([p, q])
                        V[:, pair] = mult(V[:, pair], G)
                        CM[pair, :] = mult(G.T, CM[pair, :])
                        CM[:, concatenate([Ip, Iq])]= append(c*CM[:, Ip]+
                                                             s*CM[:, Iq],
                                                             -s*CM[:, Ip]+
                                                             c*CM[:, Iq],
                                                             axis=1)
                        On = On + Gain
                        Off = Off - Gain

            if verbose:
                print("completed in %d rotations" % upds)
            updates += upds
            if updates > max_it:
                err_msg = 'No convergence after %d iterations.' % max_it
                raise mdp.NodeException(err_msg)

        if verbose:
            print("jade -> Total of %d Givens rotations" % updates)

        # A separating matrix
        # ===================
        # B is whitening matrix

        B = V.T

        # Permute the rows of the separating matrix B to get the most energetic
        # components first. Here the **signals** are normalized to unit
        # variance. Therefore, the sort is according to the norm of the
        # columns of A = pinv(B)

        if verbose:
            print("jade -> Sorting the components")

        A = numx_linalg.pinv(B)
        B =  B[numx.argsort((A*A).sum(axis=0))[::-1], :]

        if verbose:
            print("jade -> Fixing the signs")
        b = B[:, 0]
        # just a trick to deal with sign == 0
        signs = numx.sign(numx.sign(b)+0.1)
        B = mult(numx.diag(signs), B)
        self.filters = B.T
        return theta



"""
Parallel versions of hinet nodes.

Note that internal nodes are referenced instead of copied, in order to save
memory.
"""

import mdp.hinet as hinet

from . import parallelnodes


class ParallelFlowNode(hinet.FlowNode, parallelnodes.ParallelExtensionNode):
    """Parallel version of FlowNode."""

    def _fork(self):
        """Fork nodes that require it, reference all other nodes.

        If a required fork() fails the exception is not caught here.
        """   
        node_list = []
        found_train_node = False  # set to True at the first training node
        for node in self._flow:
            if not found_train_node and node.is_training():
                found_train_node = True
                node_list.append(node.fork())
            elif node.use_execute_fork():
                node_list.append(node.fork())
            else:
                node_list.append(node)
        return self.__class__(self._flow.__class__(node_list))
    
    def _join(self, forked_node):
        """Join the required nodes from the forked node into this FlowNode."""
        found_train_node = False  # set to True at the first training node          
        for i_node, node in enumerate(forked_node._flow):
            if not found_train_node and node.is_training():
                found_train_node = True
                self._flow[i_node].join(node)
            elif node.use_execute_fork():
                self._flow[i_node].join(node)
    
    def use_execute_fork(self):
        return any(node.use_execute_fork() for node in self._flow)
    

class ParallelLayer(hinet.Layer, parallelnodes.ParallelExtensionNode):
    """Parallel version of a Layer."""

    def _fork(self):
        """Fork or copy all the nodes in the layer to fork the layer."""
        forked_nodes = []
        for node in self.nodes:
            if node.is_training():
                forked_nodes.append(node.fork())
            else:
                forked_nodes.append(node)
        return self.__class__(forked_nodes)

    def _join(self, forked_node):
        """Join the trained nodes from the forked layer."""
        for i_node, layer_node in enumerate(self.nodes):
            if layer_node.is_training():
                layer_node.join(forked_node.nodes[i_node])
    
    def use_execute_fork(self):
        return any(node.use_execute_fork() for node in self.nodes)


class ParallelCloneLayer(hinet.CloneLayer, parallelnodes.ParallelExtensionNode):
    """Parallel version of CloneLayer class."""

    def _fork(self):
        """Fork the internal node in the clone layer."""
        return self.__class__(self.node.fork(), n_nodes=len(self.nodes))

    def _join(self, forked_node):
        """Join the internal node in the clone layer."""
        self.node.join(forked_node.node)
    
    def use_execute_fork(self):
        return self.node.use_execute_fork()



from __future__ import print_function
from __future__ import division
from builtins import range
from past.utils import old_div
from builtins import object
__docformat__ = "restructuredtext en"

import math
import mdp
from .isfa_nodes import ISFANode
numx, numx_rand, numx_linalg = mdp.numx, mdp.numx_rand, mdp.numx_linalg

utils = mdp.utils
mult = utils.mult

class ProjectMatrixMixin(object):
    """Mixin class to be inherited by all ICA-like algorithms"""
    def get_projmatrix(self, transposed=1):
        """Return the projection matrix."""
        self._if_training_stop_training()
        Q = self.filters.T
        if not self.whitened:
            W = self.white.get_projmatrix(transposed=0)
            T = mult(Q, W)
        else:
            T = Q
        if transposed:
            return T.T
        return T

    def get_recmatrix(self, transposed=1):
        """Return the back-projection matrix (i.e. the reconstruction matrix).
        Note that if the unknown sources are white, this is a good
        approximation of the mixing matrix (up to a permutation matrix).
        """
        self._if_training_stop_training()
        Q = self.filters.T
        if not self.whitened:
            W = self.white.get_recmatrix(transposed=1)
            T = mult(Q, W)
        else:
            T = Q
        if transposed:
            return T
        return T.T

class ICANode(mdp.Cumulator, mdp.Node, ProjectMatrixMixin):
    """
    ICANode is a general class to handle different batch-mode algorithm for
    Independent Component Analysis. More information about ICA can be found
    among others in
    Hyvarinen A., Karhunen J., Oja E. (2001). Independent Component Analysis,
    Wiley.
    """

    def __init__(self, limit = 0.001, telescope = False, verbose = False,
                 whitened = False, white_comp = None, white_parm = None,
                 input_dim = None, dtype = None):
        """
        Input arguments:

        whitened -- Set whitened is True if input data are already whitened.
                    Otherwise the node will whiten the data itself.

        white_comp -- If whitened is False, you can set 'white_comp' to the
                      number of whitened components to keep during the
                      calculation (i.e., the input dimensions are reduced to
                      white_comp by keeping the components of largest variance).

        white_parm -- a dictionary with additional parameters for whitening.
                      It is passed directly to the WhiteningNode constructor.
                      Ex: white_parm = { 'svd' : True }

        limit -- convergence threshold.

        telescope -- If telescope == True, use Telescope mode: Instead of
          using all input data in a single batch try larger and larger chunks
          of the input data until convergence is achieved. This should lead to
          significantly faster convergence for stationary statistics. This mode
          has not been thoroughly tested and must be considered beta.
        """
        self.telescope = telescope
        self.verbose = verbose
        self.limit = limit
        self.whitened = whitened
        self.white_comp = white_comp
        if white_parm is None:
            self.white_parm = {}
        else:
            self.white_parm = white_parm
        super(ICANode, self).__init__(input_dim, None, dtype)


    def _set_input_dim(self, n):
        self._input_dim = n
        if self.whitened:
            self.output_dim = n
        elif self.white_comp is None:
            self.output_dim = n

    def _stop_training(self):
        """Whiten data if needed and call the 'core' routine to perform ICA.
           Take care of telescope-mode if needed.
        """
        super(ICANode, self)._stop_training()

        verbose = self.verbose
        core = self.core
        limit = self.limit

        # ?? rewrite as a 2-phases node
        # whiten if needed
        if not self.whitened:
            self.output_dim = self.white_comp
            white = mdp.nodes.WhiteningNode(output_dim = self.white_comp,
                                            dtype=self.dtype,
                                            **self.white_parm)
            white.train(self.data)
            self.data = white.execute(self.data)
            self.white = white

        # if output_dim not set, set it now
        if self.output_dim is None:
            self.output_dim = self.input_dim

        data = self.data

        # call 'core' in telescope mode if needed
        if self.telescope:
            minpow = math.frexp(self.input_dim*10)[1]
            maxpow = int(old_div(numx.log(data.shape[0]),numx.log(2)))
            for tel in range(minpow, maxpow+1):
                index = 2**tel
                if verbose:
                    print("--\nUsing %d inputs" % index)
                convergence = core(data[:index, :])
                if convergence <= limit:
                    break
        else:
            convergence = core(data)
        if verbose:
            print("Convergence criterium: ", convergence)
        self.convergence = convergence

    def core(self, data):
        """This is the core routine of the ICANode. Each subclass must
        define this function to return the achieved convergence value.
        This function is also responsible for setting the ICA filters
        matrix self.filters.
        Note that the matrix self.filters is applied to the right of the
        matrix containing input data. This is the transposed of the matrix
        defining the linear transformation."""
        pass

    def _execute(self, x):
        if not self.whitened:
            x = self.white.execute(x)
        # self.filters is applied to the right of the
        # matrix containing input data. This is the transposed of the matrix
        # defining the linear transformation.
        return mult(x, self.filters)

    def _inverse(self, y):
        y = mult(y, self.filters.T)
        if not self.whitened:
            y = self.white.inverse(y)
        return y

class CuBICANode(ICANode):
    """
    Perform Independent Component Analysis using the CuBICA algorithm.
    Note that CuBICA is a batch-algorithm, which means that it needs
    all input data before it can start and compute the ICs.  The
    algorithm is here given as a Node for convenience, but it actually
    accumulates all inputs it receives. Remember that to avoid running
    out of memory when you have many components and many time samples.

    As an alternative to this batch mode you might consider the telescope
    mode (see the docs of the ``__init__`` method).

    Reference:
    Blaschke, T. and Wiskott, L. (2003).
    CuBICA: Independent Component Analysis by Simultaneous Third- and
    Fourth-Order Cumulant Diagonalization.
    IEEE Transactions on Signal Processing, 52(5), pp. 1250-1256.

    **Internal variables of interest**

      ``self.white``
          The whitening node used for preprocessing.

      ``self.filters``
          The ICA filters matrix (this is the transposed of the
          projection matrix after whitening).

      ``self.convergence``
          The value of the convergence threshold.
    """

    def core(self, data):
        # keep track of maximum angle of rotation
        # angles vary in the range [-pi, +pi]
        # put here -2pi < -pi < +pi
        self.maxangle = [-2*numx.pi]
        verbose = self.verbose

        # we need to copy to avoid overwriting during rotation.
        x = data.copy()

        # convergence criterium == maxangle
        limit = self.limit
        comp = x.shape[1]
        tlen = x.shape[0]

        # some constants
        ct_c34 = 0.0625
        ct_s34 = 0.25
        ct_c44 = old_div(1.,384)
        ct_s44 = old_div(1.,96)

        # initial transposed rotation matrix == identity matrix
        Qt = numx.identity(comp, dtype=self.dtype)

        # maximum number of sweeps through all possible pairs of signals
        num = int(1+round(numx.sqrt(comp)))

        # start sweeping
        for k in range(num):
            maxangle = 0
            for i in range(comp - 1):
                for j in range(i+1, comp):
                    u1 = x[:, i]
                    u2 = x[:, j]
                    sq1 = x[:, i]*x[:, i]
                    sq2 = x[:, j]*x[:, j]

                    # calculate the cumulants of 3rd and 4th order.
                    C111  = old_div(mult(sq1, u1),tlen)
                    C112  = old_div(mult(sq1, u2),tlen)
                    C122  = old_div(mult(sq2, u1),tlen)
                    C222  = old_div(mult(sq2, u2),tlen)
                    C1111 = old_div(mult(sq1, sq1),tlen) - 3.
                    C1112 = old_div(mult(sq1*u1, u2),tlen)
                    C1122 = old_div(mult(sq1, sq2),tlen) - 1.
                    C1222 = old_div(mult(sq2*u2, u1),tlen)
                    C2222 = old_div(mult(sq2, sq2),tlen) - 3.

                    c_34 = ct_c34 * (    (C111*C111+C222*C222)-
                                      3.*(C112*C112+C122*C122)-
                                      2.*(C111*C122+C112*C222)  )
                    s_34 = ct_s34 * (     C111*C112-C122*C222   )
                    c_44 = ct_c44 *(  7.*(C1111*C1111+C2222*C2222)-
                                     16.*(C1112*C1112+C1222*C1222)-
                                     12.*(C1111*C1122+C1122*C2222)-
                                     36.*(C1122*C1122)-
                                     32.*(C1112*C1222)-
                                      2.*(C1111*C2222)              )
                    s_44 = ct_s44 *(  7.*(C1111*C1112-C1222*C2222)+
                                      6.*(C1112*C1122-C1122*C1222)+
                                                (C1111*C1222-C1112*C2222)  )

                    # rotation angle that maximize the contrast function
                    phi_max = -0.25 * numx.arctan2(s_34+s_44, c_34+c_44)

                    # get the new rotation matrix.
                    # Using the function rotate with angle 'phi' on
                    # a transformation matrix corresponds to the
                    # right-multiplication by a rotation matrix
                    # with angle '-phi'.
                    utils.rotate(Qt, phi_max, [i, j])

                    # rotate input data
                    utils.rotate(x, phi_max, [i, j])

                    # keep track of maximum angle of rotation
                    maxangle = max(maxangle, abs(float(phi_max)))

            self.maxangle.append(maxangle)
            if maxangle <= limit:
                break

        self.iter = k
        if verbose:
            print("\nSweeps: ", k)
        self.filters = Qt

        # return the convergence criterium
        return maxangle

class FastICANode(ICANode):
    """
    Perform Independent Component Analysis using the FastICA algorithm.
    Note that FastICA is a batch-algorithm. This means that it needs
    all input data before it can start and compute the ICs.
    The algorithm is here given as a Node for convenience, but it
    actually accumulates all inputs it receives. Remember that to avoid
    running out of memory when you have many components and many time samples.

    FastICA does not support the telescope mode (the convergence
    criterium is not robust in telescope mode).

    Reference:
    Aapo Hyvarinen (1999).
    Fast and Robust Fixed-Point Algorithms for Independent Component Analysis
    IEEE Transactions on Neural Networks, 10(3):626-634.

    **Internal variables of interest**

      ``self.white``
          The whitening node used for preprocessing.

      ``self.filters``
          The ICA filters matrix (this is the transposed of the
          projection matrix after whitening).

      ``self.convergence``
          The value of the convergence threshold.

    History:

    - 1.4.1998 created for Matlab by Jarmo Hurri, Hugo Gavert, Jaakko Sarela,
      and Aapo Hyvarinen
    - 7.3.2003  modified for Python by Thomas Wendler
    - 3.6.2004  rewritten and adapted for scipy and MDP by MDP's authors
    - 25.5.2005 now independent from scipy. Requires Numeric or numarray
    - 26.6.2006 converted to numpy
    - 14.9.2007 updated to Matlab version 2.5
    - 26.6.2012 added ability to run two stages of optimization [PK]
    """

    def __init__(self, approach = 'defl', g = 'pow3', guess = None,
                 fine_g = 'pow3', mu = 1,
                 sample_size = 1, fine_tanh = 1, fine_gaus = 1,
                 max_it = 5000, max_it_fine = 100,
                 failures = 5, coarse_limit=None, limit = 0.001,  verbose = False,
                 whitened = False, white_comp = None, white_parm = None,
                 input_dim = None, dtype=None):
        """
        Input arguments:

        General:

        whitened -- Set whitened == True if input data are already whitened.
                    Otherwise the node will whiten the data itself

        white_comp -- If whitened == False, you can set 'white_comp' to the
                      number of whitened components to keep during the
                      calculation (i.e., the input dimensions are reduced to
                      white_comp by keeping the components of largest variance).

        white_parm -- a dictionary with additional parameters for whitening.
                      It is passed directly to the WhiteningNode constructor.
                      Ex: white_parm = { 'svd' : True }

        limit -- convergence threshold.

        Specific for FastICA:

        approach  -- Approach to use. Possible values are:
                                          'defl' --> deflation
                                          'symm' --> symmetric

               g  -- Nonlinearity to use. Possible values are:
                                          'pow3' --> x^3
                                          'tanh' --> tanh(fine_tanh*x)
                                          'gaus' --> x*exp(-fine_gaus*x^2/2)
                                          'skew' --> x^2 (for skewed signals)

           fine_g -- Nonlinearity for fine tuning. Possible values
                     are the same as for 'g'. Set it to None to disable fine
                     tuning.

               mu -- Step size. If mu != 1, a stabilization procedure is used:
                     the value of mu can momentarily be halved if the algorithm
                     is stuck between two points (this is called a stroke).
                     Also if there is no convergence before half of the maximum
                     number of iterations has been reached then mu will be halved
                     for the rest of the rounds.

      sample_size -- Percentage of samples used in one iteration. If
                     sample_size < 1, samples are chosen in random order.

     coarse_limit -- initial convergence threshold, to switch to
                     fine_g function (i.e. linear to non-linear) even
                     before reaching the limit and final tuning. Set
                     it to a value higher than limit to be in effect.

        fine_tanh -- parameter for 'tanh' nonlinearity
        fine_gaus -- parameter for 'gaus' nonlinearity

            guess -- initial guess for the mixing matrix (ignored if None)

           max_it -- maximum number of iterations

      max_it_fine -- maximum number of iterations for fine tuning

         failures -- maximum number of failures to allow in deflation mode

        """
        super(FastICANode, self).__init__(limit, False, verbose, whitened,
                                          white_comp, white_parm, input_dim,
                                          dtype)

        if approach in ['defl', 'symm']:
            self.approach = approach
        else:
            raise mdp.NodeException('%s approach method not known' % approach)

        if g in ['pow3', 'tanh', 'gaus', 'skew']:
            self.g = g
        else:
            raise mdp.NodeException('%s nonlinearity function not known' % g)

        if fine_g in ['pow3', 'tanh', 'gaus', 'skew', None]:
            self.fine_g = fine_g
        else:
            errmsg = '%s nonlinearity function not known' % fine_g
            raise mdp.NodeException(errmsg)

        if sample_size > 0 and sample_size <= 1:
            self.sample_size = sample_size
        else:
            raise mdp.NodeException('0<sample_size<1, %f given' % sample_size)

        self.mu = mu
        self.stabilization = mu != 1
        self.fine_tanh = fine_tanh
        self.fine_gaus = fine_gaus
        self.max_it = max_it
        self.max_it_fine = max_it_fine
        self.coarse_limit = coarse_limit
        self.failures = failures
        self.guess = guess

    def _get_rsamples(self, X):
        tlen = X.shape[1]
        mask = numx.where(numx_rand.random(tlen) < self.sample_size)[0]
        return X[:, mask]

    def core(self, data):
        # this is a more or less line per line translation of the original
        # matlab code.
        # Everything could be done better and more efficiently.
        # I just had no time at the moment to do it.
        # The logic behind the used_g hell is beyond my understanding :-)))

        X = data.T

        # casted constants
        comp = X.shape[0]
        tlen = X.shape[1]
        dtype = self.dtype

        # Default values and initial definitions
        fine_tanh = self.fine_tanh
        fine_gaus = self.fine_gaus
        approach = self.approach
        g = self.g
        fine_g = self.fine_g
        stabilization = self.stabilization
        mu = self.mu
        sample_size = self.sample_size
        if self.guess is None:
            # Take random orthonormal initial vectors.
            guess = utils.random_rot(comp, dtype)
        else:
            # Use user supplied mixing matrix
            guess = self._refcast(self.guess)
            if not self.whitened:
                guess = mult(guess, self.white.get_recmatrix(transposed=1))

        limit = self.limit
        coarse_limit = self.coarse_limit
        max_it = self.max_it
        max_it_fine = self.max_it_fine
        failures = self.failures
        verbose = self.verbose

        # set non linearities. don't blame me for the awful logic: it comes
        # from the matlab program. I didn't dare to understand it and change
        # it.
        if g == 'pow3':
            gOrig = 10
        elif g == 'tanh':
            gOrig = 20
        elif g == 'gaus':
            gOrig = 30
        else:
            gOrig = 40

        if sample_size != 1:
            gOrig += 2
        if mu != 1:
            gOrig += 1

        fine_tuning = True
        if fine_g == 'pow3':
            gFine = 11
        elif fine_g == 'tanh':
            gFine = 21
        elif fine_g == 'gaus':
            gFine = 31
        elif fine_g == 'skew':
            gFine = 41
        else:
            if mu == 1:
                gFine = gOrig + 1
            else:
                stabilization = True
                gFine = gOrig
            fine_tuning = False

        muK = 0.01
        used_g = gOrig
        stroke = 0
        fine_tuned = False
        coarse_limit_reached = False
        lng = False

        # SYMMETRIC APPROACH
        if approach == 'symm':
            # create list to store convergence
            convergence = []
            convergence_fine = []
            # orthonormal initial vectors.
            Q = guess
            QOld = numx.zeros(Q.shape, dtype)
            QOldF = numx.zeros(Q.shape, dtype)
            # This is the actual fixed-point iteration loop.
            for round in range(max_it + 1):
                if round == max_it:
                    errstr = 'No convergence after %d steps\n' % max_it
                    raise mdp.NodeException(errstr)


                # Symmetric orthogonalization. Q = Q * real(inv(Q' * Q)^(1/2));
                Q = mult(Q, utils.sqrtm(utils.inv(mult(Q.T, Q))))

                # Test for termination condition. Note that we consider
                # opposite directions here as well.
                v1 = 1.-abs((mult(Q.T, QOld)).diagonal()).min(axis=0)
                convergence.append(v1)
                v2 = 1.-abs((mult(Q.T, QOldF)).diagonal()).min(axis=0)
                convergence_fine.append(v2)

                if self.g != self.fine_g \
                   and coarse_limit is not None \
                   and convergence[round] < coarse_limit \
                   and not coarse_limit_reached:
                    if verbose:
                        print('Coarse convergence, switching to fine cost...')
                    used_g = gFine
                    coarse_limit_reached = True

                if convergence[round] < limit:
                    if fine_tuning and (not fine_tuned):
                        if verbose:
                            print('Initial convergence, fine-tuning...')
                        fine_tuned = True
                        used_g = gFine
                        mu = muK * self.mu
                        QOld = numx.zeros(Q.shape, dtype)
                        QoldF = numx.zeros(Q.shape, dtype)
                    else:
                        if verbose:
                            print('Convergence after %d steps\n' % round)
                        break
                if stabilization:
                    if (stroke == 0) and (convergence_fine[round] < limit):
                        if verbose:
                            print('Stroke!\n')
                        stroke = mu
                        mu = 0.5*mu
                        if used_g % 2 == 0:
                            used_g += 1
                    elif (stroke != 0):
                        mu = stroke
                        stroke = 0
                        if (mu == 1) and (used_g % 2 != 0):
                            used_g -= 1
                    elif (not lng) and (round > max_it//2):
                        if verbose:
                            print('Taking long (reducing step size)...')
                        lng = True
                        mu = 0.5*mu
                        if used_g % 2 == 0:
                            used_g += 1

                QOldF = QOld
                QOld = Q

                # Show the progress...
                if verbose:
                    msg = ('Step no. %d,'
                           ' convergence: %.7f' % (round+1,convergence[round]))
                    print(msg)


                # First calculate the independent components (u_i's).
                # u_i = b_i' x = x' b_i. For all x:s simultaneously this is
                # non linearity
                if used_g == 10:
                    u = mult(X.T, Q)
                    Q = old_div(mult(X, u*u*u),tlen) - 3.*Q
                elif used_g == 11:
                    u = mult(X.T, Q)
                    Gpow3 = u*u*u
                    Beta = (u*Gpow3).sum(axis=0)
                    D = numx.diag((old_div(1,(Beta - 3*tlen))))
                    Q = Q + mu * mult(Q, mult((mult(u.T, Gpow3) -
                                               numx.diag(Beta)), D))
                elif used_g == 12:
                    Xsub = self._get_rsamples(X)
                    u = mult(Xsub.T, Q)
                    Q = old_div(mult(Xsub, u*u*u),Xsub.shape[1]) - 3.*Q
                elif used_g == 13:
                    Xsub = self._get_rsamples(X)
                    u = mult(Xsub.T, Q)
                    Gpow3 = u*u*u
                    Beta = (u*Gpow3).sum(axis=0)
                    D = numx.diag((old_div(1,(Beta - 3*Xsub.shape[1]))))
                    Q = Q + mu * mult(Q, mult((mult(u.T, Gpow3) -
                                               numx.diag(Beta)), D))
                elif used_g == 20:
                    u = mult(X.T, Q)
                    tang = numx.tanh(fine_tanh * u)
                    temp = old_div((1.-tang*tang).sum(axis=0),tlen)
                    Q = old_div(mult(X, tang),tlen) - temp * Q * fine_tanh
                elif used_g == 21:
                    u = mult(X.T, Q)
                    tang = numx.tanh(fine_tanh * u)
                    Beta = (u*tang).sum(axis=0)
                    D = numx.diag(old_div(1,(Beta -
                                     fine_tanh*(1.-tang*tang).sum(axis=0))))
                    Q = Q + mu * mult(Q,
                                 mult((mult(u.T, tang)-
                                       numx.diag(Beta)), D))
                elif used_g == 22:
                    Xsub = self._get_rsamples(X)
                    u = mult(Xsub.T, Q)
                    tang = numx.tanh(fine_tanh * u)
                    temp = old_div((1.-tang*tang).sum(axis=0),Xsub.shape[1])
                    Q = old_div(mult(Xsub, tang),Xsub.shape[1]) - temp * Q * fine_tanh
                elif used_g == 23:
                    Xsub = self._get_rsamples(X)
                    u = mult(Xsub.T, Q)
                    tang = numx.tanh(fine_tanh * u)
                    Beta = (u*tang).sum(axis=0)
                    D = numx.diag(old_div(1,(Beta -
                                     fine_tanh*(1.-tang*tang).sum(axis=0))))
                    Q = Q + mu * mult(Q,
                                 mult((mult(u.T, tang)-
                                       numx.diag(Beta)), D))
                elif used_g == 30:
                    u = mult(X.T, Q)
                    u2 = u*u
                    ex = numx.exp(-fine_gaus*u2*0.5)
                    gauss =  u*ex
                    dgauss = (1. - fine_gaus*u2)*ex
                    Q = old_div((mult(X, gauss)-dgauss.sum(axis=0)*Q),tlen)
                elif used_g == 31:
                    u = mult(X.T, Q)
                    u2 = u*u
                    ex = numx.exp(-fine_gaus*u2*0.5)
                    gaus =  u*ex
                    Beta = (u*gaus).sum(axis=0)
                    D = numx.diag(old_div(1,(Beta -
                                     ((1-fine_gaus*u2)*ex).sum(axis=0))))
                    Q = Q + mu * mult(Q,
                                 mult((mult(u.T, gaus)-
                                       numx.diag(Beta)), D))
                elif used_g == 32:
                    Xsub = self._get_rsamples(X)
                    u = mult(Xsub.T, Q)
                    u2 = u*u
                    ex = numx.exp(-fine_gaus*u2*0.5)
                    gauss =  u*ex
                    dgauss = (1. - fine_gaus*u2)*ex
                    Q = old_div((mult(Xsub, gauss)-dgauss.sum(axis=0)*Q),Xsub.shape[1])
                elif used_g == 33:
                    Xsub = self._get_rsamples(X)
                    u = mult(Xsub.T, Q)
                    u2 = u*u
                    ex = numx.exp(-fine_gaus*u2*0.5)
                    gaus = u*ex
                    Beta = (u*gaus).sum(axis=0)
                    D = numx.diag(old_div(1,(Beta -
                                     ((1-fine_gaus*u2)*ex).sum(axis=0))))
                    Q = Q + mu * mult(Q, mult((mult(u.T, gaus)-
                                               numx.diag(Beta)), D))
                elif used_g == 40:
                    u = mult(X.T, Q)
                    Q = old_div(mult(X, u*u),tlen)
                elif used_g == 41:
                    u = mult(X.T, Q)
                    Gskew = u*u
                    Beta = (u*Gskew).sum(axis=0)
                    D = numx.diag(old_div(1,Beta))
                    Q =  Q + mu * mult(Q, mult((mult(u.T, Gskew)-
                                                numx.diag(Beta)), D))
                elif used_g == 42:
                    Xsub = self._get_rsamples(X)
                    u = mult(Xsub.T, Q)
                    Q = old_div(mult(Xsub, u*u),Xsub.shape[1])
                elif used_g == 43:
                    Xsub = self._get_rsamples(X)
                    u = mult(Xsub.T, Q)
                    Gskew = u*u
                    Beta = (u*Gskew).sum(axis=0)
                    D = numx.diag(old_div(1,Beta))
                    Q =  Q + mu * mult(Q, mult((mult(u.T, Gskew)-
                                                numx.diag(Beta)), D))
                else:
                    errstr = 'Nonlinearity not found: %i' % used_g
                    raise mdp.NodeException(errstr)

            self.convergence = numx.array(convergence)
            self.convergence_fine = numx.array(convergence_fine)
            ret = convergence[-1]
        # DEFLATION APPROACH
        elif approach == 'defl':
            # adjust limit!
            #limit = 1 - limit*limit*0.5
            # create array to store convergence
            convergence = []
            convergence_fine = []
            Q = numx.zeros((comp, comp), dtype=dtype)
            round = 0
            nfail = 0
            while round < comp:
                mu = self.mu
                used_g = gOrig
                stroke = 0
                fine_tuned = False
                lng = False
                end_finetuning = 0

                # Take a random initial vector of lenght 1 and orthogonalize it
                # with respect to the other vectors.
                w  = guess[:, round]
                w -= mult(mult(Q, Q.T), w)
                w /= utils.norm2(w)

                wOld = numx.zeros(w.shape, dtype)
                wOldF = numx.zeros(w.shape, dtype)
                # This is the actual fixed-point iteration loop.
                i = 1
                gabba = 1
                #for i in range(max_it + 1):
                while i <= max_it + gabba:
                    # Project the vector into the space orthogonal to the space
                    # spanned by the earlier found basis vectors. Note that
                    # we can do the projection with matrix Q, since the zero
                    # entries do not contribute to the projection.
                    w -= mult(mult(Q, Q.T), w)
                    w /= utils.norm2(w)

                    if not fine_tuned:
                        if i == max_it + 1:
                            err_msg = ('Component number %d did not'
                                       'converge in %d iterations.' % (round,
                                                                       max_it))
                            if verbose:
                                print(err_msg)
                            if round == 0:
                                raise mdp.NodeException(err_msg)
                            nfail += 1
                            if nfail > failures:
                                err = ('Too many failures to '
                                       'converge (%d). Giving up.' % nfail)
                                raise mdp.NodeException(err)
                            break
                    else:
                        if i >= end_finetuning:
                            wOld = w

                    # Test for termination condition. Note that the algorithm
                    # has converged if the direction of w and wOld is the same.
                    #conv = float(abs((w*wOld).sum()))
                    conv = min(utils.norm2(w-wOld), utils.norm2(w+wOld))
                    convergence.append(conv)
                    if conv < limit:
                        if fine_tuning and (not fine_tuned):
                            if verbose:
                                print('Initial convergence, fine-tuning...')
                            fine_tuned = True
                            gabba = max_it_fine
                            wOld = numx.zeros(w.shape, dtype)
                            wOldF = numx.zeros(w.shape, dtype)
                            used_g = gFine
                            mu = muK * self.mu
                            end_finetuning = max_it_fine + i
                        else:
                            nfail = 0
                            convergence[round] = conv
                            # Calculate ICA filter.
                            Q[:, round] = w.copy()
                            # Show the progress...
                            if verbose:
                                print('IC %d computed ( %d steps )' % (round+1,
                                                                       i+1))
                            break
                    elif stabilization:
                        conv_fine = min(utils.norm2(w-wOldF),
                                        utils.norm2(w+wOldF))
                        convergence_fine.append(conv_fine)
                        if  (stroke == 0) and conv_fine < limit:
                            if verbose:
                                print('Stroke!')
                            stroke = mu
                            mu = 0.5*mu
                            if used_g % 2 == 0:
                                used_g += 1
                        elif (stroke != 0):
                            mu = stroke
                            stroke = 0
                            if (mu == 1) and (used_g % 2 != 0):
                                used_g -= 1
                        elif (not lng) and (i > max_it//2):
                            if verbose:
                                print('Taking long (reducing step size)...')
                            lng = True
                            mu = 0.5*mu
                            if used_g % 2 == 0:
                                used_g += 1

                    wOldF = wOld
                    wOld = w
                    if used_g == 10:
                        u = mult(X.T, w)
                        w = old_div(mult(X, u*u*u),tlen) - 3.*w
                    elif used_g == 11:
                        u = mult(X.T, w)
                        EXGpow3 = old_div(mult(X, u*u*u),tlen)
                        Beta = mult(w.T, EXGpow3)
                        w = w - mu * (EXGpow3 - Beta*w)/(3-Beta)
                    elif used_g == 12:
                        Xsub = self._get_rsamples(X)
                        u = mult(Xsub.T, w)
                        w = old_div(mult(Xsub, u*u*u),Xsub.shape[1]) - 3.*w
                    elif used_g == 13:
                        Xsub = self._get_rsamples(X)
                        u = mult(Xsub.T, w)
                        EXGpow3 = old_div(mult(Xsub, u*u*u),Xsub.shape[1])
                        Beta = mult(w.T, EXGpow3)
                        w = w - mu * (EXGpow3 - Beta*w)/(3-Beta)
                    elif used_g == 20:
                        u = mult(X.T, w)
                        tang = numx.tanh(fine_tanh * u)
                        temp = mult((1. - tang*tang).sum(axis=0), w)
                        w = old_div((mult(X, tang) - fine_tanh*temp),tlen)
                    elif used_g == 21:
                        u = mult(X.T, w)
                        tang = numx.tanh(fine_tanh * u)
                        Beta = mult(u.T, tang)
                        temp = (1. - tang*tang).sum(axis=0)
                        w = w-mu*(old_div((mult(X, tang)-Beta*w),(fine_tanh*temp-Beta)))
                    elif used_g == 22:
                        Xsub = self._get_rsamples(X)
                        u = mult(Xsub.T, w)
                        tang = numx.tanh(fine_tanh * u)
                        temp = mult((1. - tang*tang).sum(axis=0), w)
                        w = old_div((mult(Xsub, tang) - fine_tanh*temp),Xsub.shape[1])
                    elif used_g == 23:
                        Xsub = self._get_rsamples(X)
                        u = mult(Xsub.T, w)
                        tang = numx.tanh(fine_tanh * u)
                        Beta = mult(u.T, tang)
                        w = w - mu * (old_div((mult(Xsub, tang)-Beta*w),
                                      (fine_tanh*(1. - tang*tang).sum(axis=0) -
                                       Beta)))
                    elif used_g == 30:
                        u = mult(X.T, w)
                        u2 = u*u
                        ex = numx.exp(-fine_gaus*u2*0.5)
                        gauss =  u*ex
                        dgauss = (1. - fine_gaus *u2)*ex
                        w = old_div((mult(X, gauss)-mult(dgauss.sum(axis=0), w)),tlen)
                    elif used_g == 31:
                        u = mult(X.T, w)
                        u2 = u*u
                        ex = numx.exp(-fine_gaus*u2*0.5)
                        gauss =  u*ex
                        dgauss = (1. - fine_gaus *u2)*ex
                        Beta = mult(u.T, gauss)
                        w = w - mu*(old_div((mult(X, gauss)-Beta*w),
                                    (dgauss.sum(axis=0)-Beta)))
                    elif used_g == 32:
                        Xsub = self._get_rsamples(X)
                        u = mult(Xsub.T, w)
                        u2 = u*u
                        ex = numx.exp(-fine_gaus*u2*0.5)
                        gauss =  u*ex
                        dgauss = (1. - fine_gaus *u2)*ex
                        w = old_div((mult(Xsub, gauss)-
                             mult(dgauss.sum(axis=0), w)),Xsub.shape[1])
                    elif used_g == 33:
                        Xsub = self._get_rsamples(X)
                        u = mult(Xsub.T, w)
                        u2 = u*u
                        ex = numx.exp(-fine_gaus*u2*0.5)
                        gauss =  u*ex
                        dgauss = (1. - fine_gaus *u2)*ex
                        Beta = mult(u.T, gauss)
                        w = w - mu*(old_div((mult(Xsub, gauss)-Beta*w),
                                    (dgauss.sum(axis=0)-Beta)))
                    elif used_g == 40:
                        u = mult(X.T, w)
                        w = old_div(mult(X, u*u),tlen)
                    elif used_g == 41:
                        u = mult(X.T, w)
                        EXGskew = old_div(mult(X, u*u), tlen)
                        Beta = mult(w.T, EXGskew)
                        w = w - mu * (EXGskew - mult(Beta, w))/(-Beta)
                    elif used_g == 42:
                        Xsub = self._get_rsamples(X)
                        u = mult(Xsub.T, w)
                        w = old_div(mult(Xsub, u*u),Xsub.shape[1])
                    elif used_g == 43:
                        Xsub = self._get_rsamples(X)
                        u = mult(Xsub.T, w)
                        EXGskew = old_div(mult(Xsub, u*u), Xsub.shape[1])
                        Beta = mult(w.T, EXGskew)
                        w = w - mu * (EXGskew - Beta*w)/(-Beta)
                    else:
                        errstr = 'Nonlinearity not found: %i' % used_g
                        raise mdp.NodeException(errstr)

                    # Normalize the new w.
                    w /= utils.norm2(w)
                    i += 1

                round += 1
            self.convergence = numx.array(convergence)
            self.convergence_fine = numx.array(convergence_fine)
            ret = convergence[-1]
        self.filters = Q
        return ret


class TDSEPNode(ISFANode, ProjectMatrixMixin):
    """Perform Independent Component Analysis using the TDSEP algorithm.
    Note that TDSEP, as implemented in this Node, is an online algorithm,
    i.e. it is suited to be trained on huge data sets, provided that the
    training is done sending small chunks of data for each time.

    Reference:
    Ziehe, Andreas and Muller, Klaus-Robert (1998).
    TDSEP an efficient algorithm for blind separation using time structure.
    in Niklasson, L, Boden, M, and Ziemke, T (Editors), Proc. 8th Int. Conf.
    Artificial Neural Networks (ICANN 1998).

    **Internal variables of interest**

      ``self.white``
          The whitening node used for preprocessing.

      ``self.filters``
          The ICA filters matrix (this is the transposed of the
          projection matrix after whitening).

      ``self.convergence``
          The value of the convergence threshold.
    """
    def __init__(self, lags=1, limit = 0.00001, max_iter=10000,
                 verbose = False, whitened = False, white_comp = None,
                 white_parm = None, input_dim = None, dtype = None):
        """
        Input arguments:

        lags    -- list of time-lags to generate the time-delayed covariance
                   matrices. If lags is an integer, time-lags 1,2,...,'lags'
                   are used.
                   Note that time-lag == 0 (instantaneous correlation) is
                   always implicitly used.

        whitened -- Set whitened is True if input data are already whitened.
                    Otherwise the node will whiten the data itself.

        white_comp -- If whitened is False, you can set 'white_comp' to the
                      number of whitened components to keep during the
                      calculation (i.e., the input dimensions are reduced to
                      white_comp by keeping the components of largest variance).

        white_parm -- a dictionary with additional parameters for whitening.
                      It is passed directly to the WhiteningNode constructor.
                      Ex: white_parm = { 'svd' : True }

        limit -- convergence threshold.

        max_iter     -- If the algorithms does not achieve convergence within
                        max_iter iterations raise an Exception. Should be
                        larger than 100.
        """
        super(TDSEPNode, self).__init__(lags=lags, sfa_ica_coeff=(0., 1.),
                                        icaweights=None, sfaweights=None,
                                        whitened=whitened,
                                        white_comp=white_comp,
                                        white_parm = None,
                                        eps_contrast=limit,
                                        max_iter=max_iter, RP=None,
                                        verbose=verbose,
                                        input_dim=input_dim,
                                        output_dim=None,
                                        dtype=dtype)

    def _stop_training(self, covs=None):
        super(TDSEPNode, self)._stop_training(covs)
        # set filters
        self.filters = self.RP
        # set convergence
        self.convergence = self.final_contrast



"""Hierarchical Networks Package.

This package makes it possible to construct graph-like Node structures,
especially hierarchical networks.

The most important building block is the new Layer node, which works as an
horizontal version of flow. It encapsulates a list of Nodes, which are trained
and executed in parallel.
For example we can take two Nodes with 100 dimensional input to
construct a layer with a 200 dimensional input. The first half of the input
data is automatically fed into the first Node, the second half into the second
Node.

Since one might also want to use Flows (i.e. vertical stacks of Nodes) in a
Layer, a wrapper class for Nodes is provided.
The FlowNode class wraps any Flow into a Node, which can then be used like any
other Node. Together with the Layer this allows you to combine Nodes both
horizontally and vertically. Thereby one can in principle realize
any feed-forward network topology.

For a hierarchical networks one might want to route the different parts of the
data to different Nodes in a Layer in complicated ways. This is done by a
Switchboard that handles all the routing.
Defining the routing manually can be quite tedious, so one can derive subclasses
for special routing situations. One such subclass for 2d image data is provided.
It maps the data according to rectangular overlapping 2d input areas. One can
then feed the output into a Layer and each Node will get the correct input.
"""

from .flownode import FlowNode
from .layer import Layer, SameInputLayer, CloneLayer
from .switchboard import (
    Switchboard, SwitchboardException, MeanInverseSwitchboard,
    ChannelSwitchboard,
    Rectangular2dSwitchboard, Rectangular2dSwitchboardException,
    DoubleRect2dSwitchboard, DoubleRect2dSwitchboardException,
    DoubleRhomb2dSwitchboard, DoubleRhomb2dSwitchboardException
)
from .htmlvisitor import (
    HiNetHTMLVisitor, HiNetXHTMLVisitor, NewlineWriteFile, show_flow
)
from .switchboard_factory import (
    get_2d_image_switchboard, FactoryExtensionChannelSwitchboard,
    FactoryRectangular2dSwitchboard, FactoryDoubleRect2dSwitchboard,
    FactoryDoubleRhomb2dSwitchboard
)

__all__ = ['FlowNode', 'Layer', 'SameInputLayer', 'CloneLayer',
           'Switchboard', 'SwitchboardException', 'ChannelSwitchboard',
           'Rectangular2dSwitchboard', 'Rectangular2dSwitchboardException',
           'DoubleRect2dSwitchboard', 'DoubleRect2dSwitchboardException',
           'DoubleRhomb2dSwitchboard', 'DoubleRhomb2dSwitchboardException',
           'HiNetHTMLVisitor', 'HiNetXHTMLVisitor', 'NewlineWriteFile',
           'show_flow', 'get_2d_image_switchboard'
           ]

from mdp.utils import fixup_namespace
fixup_namespace(__name__, __all__,
                ('flownode',
                 'layer',
                 'switchboard',
                 'hinet_Visitor',
                 'switchboard_factory',
                 'utils',
                 'fixup_namespace'
                 ))
